# This file is part of cloud-init. See LICENSE file for license information.

import functools
import os
from unittest import mock

import responses

from cloudinit import helpers
from cloudinit.sources import DataSourceAliYun as ay
from cloudinit.sources.DataSourceEc2 import convert_ec2_metadata_network_config
from tests.unittests import helpers as test_helpers

DEFAULT_METADATA = {
    "instance-id": "aliyun-test-vm-00",
    "eipv4": "10.0.0.1",
    "hostname": "test-hostname",
    "image-id": "m-test",
    "launch-index": "0",
    "mac": "00:16:3e:00:00:00",
    "network-type": "vpc",
    "private-ipv4": "192.168.0.1",
    "serial-number": "test-string",
    "vpc-cidr-block": "192.168.0.0/16",
    "vpc-id": "test-vpc",
    "vswitch-id": "test-vpc",
    "vswitch-cidr-block": "192.168.0.0/16",
    "zone-id": "test-zone-1",
    "ntp-conf": {
        "ntp_servers": [
            "ntp1.aliyun.com",
            "ntp2.aliyun.com",
            "ntp3.aliyun.com",
        ]
    },
    "source-address": [
        "http://mirrors.aliyun.com",
        "http://mirrors.aliyuncs.com",
    ],
    "public-keys": {
        "key-pair-1": {"openssh-key": "ssh-rsa AAAAB3..."},
        "key-pair-2": {"openssh-key": "ssh-rsa AAAAB3..."},
    },
}

DEFAULT_USERDATA = """\
#cloud-config

hostname: localhost"""


class TestAliYunDatasource(test_helpers.ResponsesTestCase):
    def setUp(self):
        super(TestAliYunDatasource, self).setUp()
        cfg = {"datasource": {"AliYun": {"timeout": "1", "max_wait": "1"}}}
        distro = {}
        paths = helpers.Paths({"run_dir": self.tmp_dir()})
        self.ds = ay.DataSourceAliYun(cfg, distro, paths)
        self.metadata_address = self.ds.metadata_urls[0]

    @property
    def default_metadata(self):
        return DEFAULT_METADATA

    @property
    def default_userdata(self):
        return DEFAULT_USERDATA

    @property
    def metadata_url(self):
        return (
            os.path.join(
                self.metadata_address,
                self.ds.min_metadata_version,
                "meta-data",
            )
            + "/"
        )

    @property
    def userdata_url(self):
        return os.path.join(
            self.metadata_address, self.ds.min_metadata_version, "user-data"
        )

    # EC2 provides an instance-identity document which must return 404 here
    # for this test to pass.
    @property
    def default_identity(self):
        return {}

    @property
    def identity_url(self):
        return os.path.join(
            self.metadata_address,
            self.ds.min_metadata_version,
            "dynamic",
            "instance-identity",
        )

    @property
    def token_url(self):
        return os.path.join(
            self.metadata_address,
            "latest",
            "api",
            "token",
        )

    def register_mock_metaserver(self, base_url, data):
        def register_helper(register, base_url, body):
            if isinstance(body, str):
                register(base_url, body)
            elif isinstance(body, list):
                register(base_url.rstrip("/"), "\n".join(body) + "\n")
            elif isinstance(body, dict):
                if not body:
                    register(
                        base_url.rstrip("/") + "/", "not found", status=404
                    )
                vals = []
                for k, v in body.items():
                    if isinstance(v, (str, list)):
                        suffix = k.rstrip("/")
                    else:
                        suffix = k.rstrip("/") + "/"
                    vals.append(suffix)
                    url = base_url.rstrip("/") + "/" + suffix
                    register_helper(register, url, v)
                register(base_url, "\n".join(vals) + "\n")

        register = functools.partial(self.responses.add, responses.GET)
        register_helper(register, base_url, data)

    def regist_default_server(self):
        self.register_mock_metaserver(self.metadata_url, self.default_metadata)
        self.register_mock_metaserver(self.userdata_url, self.default_userdata)
        self.register_mock_metaserver(self.identity_url, self.default_identity)
        self.responses.add(responses.PUT, self.token_url, "API-TOKEN")

    def _test_get_data(self):
        self.assertEqual(self.ds.metadata, self.default_metadata)
        self.assertEqual(
            self.ds.userdata_raw, self.default_userdata.encode("utf8")
        )

    def _test_get_sshkey(self):
        pub_keys = [
            v["openssh-key"]
            for (_, v) in self.default_metadata["public-keys"].items()
        ]
        self.assertEqual(self.ds.get_public_ssh_keys(), pub_keys)

    def _test_get_iid(self):
        self.assertEqual(
            self.default_metadata["instance-id"], self.ds.get_instance_id()
        )

    def _test_host_name(self):
        self.assertEqual(
            self.default_metadata["hostname"], self.ds.get_hostname().hostname
        )

    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    def test_with_mock_server(self, m_is_aliyun, m_resolv):
        m_is_aliyun.return_value = True
        self.regist_default_server()
        ret = self.ds.get_data()
        self.assertEqual(True, ret)
        self.assertEqual(1, m_is_aliyun.call_count)
        self._test_get_data()
        self._test_get_sshkey()
        self._test_get_iid()
        self._test_host_name()
        self.assertEqual("aliyun", self.ds.cloud_name)
        self.assertEqual("ec2", self.ds.platform)
        self.assertEqual(
            "metadata (http://100.100.100.200)", self.ds.subplatform
        )

    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv6Network")
    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv4Network")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    @mock.patch("cloudinit.net.find_fallback_nic")
    @mock.patch("cloudinit.net.ephemeral.maybe_perform_dhcp_discovery")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_FreeBSD")
    def test_aliyun_local_with_mock_server(
        self,
        m_is_bsd,
        m_dhcp,
        m_fallback_nic,
        m_is_aliyun,
        m_resolva,
        m_net4,
        m_net6,
    ):
        m_is_aliyun.return_value = True
        m_fallback_nic.return_value = "eth9"
        m_dhcp.return_value = {
            "interface": "eth9",
            "fixed-address": "192.168.2.9",
            "routers": "192.168.2.1",
            "subnet-mask": "255.255.255.0",
            "broadcast-address": "192.168.2.255",
        }
        m_is_bsd.return_value = False
        cfg = {"datasource": {"AliYun": {"timeout": "1", "max_wait": "1"}}}
        distro = mock.MagicMock()
        paths = helpers.Paths({"run_dir": self.tmp_dir()})
        self.ds = ay.DataSourceAliYunLocal(cfg, distro, paths)
        self.regist_default_server()
        ret = self.ds.get_data()
        self.assertEqual(True, ret)
        self.assertEqual(1, m_is_aliyun.call_count)
        self._test_get_data()
        self._test_get_sshkey()
        self._test_get_iid()
        self._test_host_name()
        self.assertEqual("aliyun", self.ds.cloud_name)
        self.assertEqual("ec2", self.ds.platform)
        self.assertEqual(
            "metadata (http://100.100.100.200)", self.ds.subplatform
        )

    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    def test_returns_false_when_not_on_aliyun(self, m_is_aliyun):
        """If is_aliyun returns false, then get_data should return False."""
        m_is_aliyun.return_value = False
        self.regist_default_server()
        ret = self.ds.get_data()
        self.assertEqual(1, m_is_aliyun.call_count)
        self.assertEqual(False, ret)

    def test_parse_public_keys(self):
        public_keys = {}
        self.assertEqual(ay.parse_public_keys(public_keys), [])

        public_keys = {"key-pair-0": "ssh-key-0"}
        self.assertEqual(
            ay.parse_public_keys(public_keys), [public_keys["key-pair-0"]]
        )

        public_keys = {"key-pair-0": "ssh-key-0", "key-pair-1": "ssh-key-1"}
        self.assertEqual(
            set(ay.parse_public_keys(public_keys)),
            set([public_keys["key-pair-0"], public_keys["key-pair-1"]]),
        )

        public_keys = {"key-pair-0": ["ssh-key-0", "ssh-key-1"]}
        self.assertEqual(
            ay.parse_public_keys(public_keys), public_keys["key-pair-0"]
        )

        public_keys = {"key-pair-0": {"openssh-key": []}}
        self.assertEqual(ay.parse_public_keys(public_keys), [])

        public_keys = {"key-pair-0": {"openssh-key": "ssh-key-0"}}
        self.assertEqual(
            ay.parse_public_keys(public_keys),
            [public_keys["key-pair-0"]["openssh-key"]],
        )

        public_keys = {
            "key-pair-0": {"openssh-key": ["ssh-key-0", "ssh-key-1"]}
        }
        self.assertEqual(
            ay.parse_public_keys(public_keys),
            public_keys["key-pair-0"]["openssh-key"],
        )

    def test_route_metric_calculated_without_device_number(self):
        """Test that route-metric code works without `device-number`

        `device-number` is part of EC2 metadata, but not supported on aliyun.
        Attempting to access it will raise a KeyError.

        LP: #1917875
        """
        netcfg = convert_ec2_metadata_network_config(
            {
                "interfaces": {
                    "macs": {
                        "06:17:04:d7:26:09": {
                            "interface-id": "eni-e44ef49e",
                        },
                        "06:17:04:d7:26:08": {
                            "interface-id": "eni-e44ef49f",
                        },
                    }
                }
            },
            mock.Mock(),
            macs_to_nics={
                "06:17:04:d7:26:09": "eth0",
                "06:17:04:d7:26:08": "eth1",
            },
        )

        met0 = netcfg["ethernets"]["eth0"]["dhcp4-overrides"]["route-metric"]
        met1 = netcfg["ethernets"]["eth1"]["dhcp4-overrides"]["route-metric"]

        # route-metric numbers should be 100 apart
        assert 100 == abs(met0 - met1)

        # No policy routing
        assert not {"routing-policy", "routes"}.intersection(
            netcfg["ethernets"]["eth0"].keys()
        )
        assert not {"routing-policy", "routes"}.intersection(
            netcfg["ethernets"]["eth1"].keys()
        )


class TestIsAliYun(test_helpers.CiTestCase):
    ALIYUN_PRODUCT = "Alibaba Cloud ECS"
    read_dmi_data_expected = [mock.call("system-product-name")]

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_true_on_aliyun_product(self, m_read_dmi_data):
        """Should return true if the dmi product data has expected value."""
        m_read_dmi_data.return_value = self.ALIYUN_PRODUCT
        ret = ay._is_aliyun()
        self.assertEqual(
            self.read_dmi_data_expected, m_read_dmi_data.call_args_list
        )
        self.assertEqual(True, ret)

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_false_on_empty_string(self, m_read_dmi_data):
        """Should return false on empty value returned."""
        m_read_dmi_data.return_value = ""
        ret = ay._is_aliyun()
        self.assertEqual(
            self.read_dmi_data_expected, m_read_dmi_data.call_args_list
        )
        self.assertEqual(False, ret)

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_false_on_unknown_string(self, m_read_dmi_data):
        """Should return false on an unrelated string."""
        m_read_dmi_data.return_value = "cubs win"
        ret = ay._is_aliyun()
        self.assertEqual(
            self.read_dmi_data_expected, m_read_dmi_data.call_args_list
        )
        self.assertEqual(False, ret)



# This file is part of cloud-init. See LICENSE file for license information.

import errno
import gzip
import json
import os
from collections import namedtuple
from io import BytesIO
from pathlib import Path
from textwrap import dedent

import pytest

from cloudinit.atomic_helper import b64e
from cloudinit.cmd import query
from cloudinit.helpers import Paths
from cloudinit.sources import REDACT_SENSITIVE_VALUE
from cloudinit.templater import JinjaSyntaxParsingException
from cloudinit.util import write_file
from tests.unittests.helpers import mock

M_PATH = "cloudinit.cmd.query."


def _gzip_data(data):
    with BytesIO() as iobuf:
        with gzip.GzipFile(mode="wb", fileobj=iobuf, mtime=0) as gzfp:
            gzfp.write(data)
        return iobuf.getvalue()


@pytest.fixture(autouse=True)
def setup_mocks(mocker):
    mocker.patch("cloudinit.cmd.query.read_cfg_paths", return_value=Paths({}))


class TestQuery:
    Args = namedtuple(
        "Args",
        "debug dump_all format instance_data list_keys user_data vendor_data"
        " varname",
    )

    def _setup_paths(self, tmpdir, ud_val=None, vd_val=None):
        """Write userdata and vendordata into a tmpdir.

        Return:
            4-tuple : (paths, run_dir_path, userdata_path, vendordata_path)
        """
        if ud_val:
            user_data = tmpdir.join("user-data")
            write_file(user_data.strpath, ud_val)
        else:
            user_data = None
        if vd_val:
            vendor_data = tmpdir.join("vendor-data")
            write_file(vendor_data.strpath, vd_val)
        else:
            vendor_data = None
        run_dir = tmpdir.join("run_dir")
        run_dir.ensure_dir()

        cloud_dir = tmpdir.join("cloud_dir")
        cloud_dir.ensure_dir()

        return (
            Paths(
                {"cloud_dir": cloud_dir.strpath, "run_dir": run_dir.strpath}
            ),
            run_dir,
            user_data,
            vendor_data,
        )

    def test_handle_args_error_on_missing_param(self, caplog, capsys):
        """Error when missing required parameters and print usage."""
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=None,
        )
        assert 1 == query.handle_args("anyname", args)
        expected_error = (
            "Expected one of the options: --all, --format, --list-keys"
            " or varname\n"
        )
        assert expected_error in caplog.text
        out, _err = capsys.readouterr()
        assert "usage: query" in out

    @pytest.mark.parametrize(
        "inst_data,varname,expected_error",
        (
            (
                '{"v1": {"key-2": "value-2"}}',
                "v1.absent_leaf",
                "instance-data 'v1' has no 'absent_leaf'\n",
            ),
            (
                '{"v1": {"key-2": "value-2"}}',
                "absent_key",
                "Undefined instance-data key 'absent_key'\n",
            ),
        ),
    )
    def test_handle_args_error_on_invalid_varname_paths(
        self, inst_data, varname, expected_error, caplog, tmpdir
    ):
        """Error when varname is not a valid instance-data variable path."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(inst_data)
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=varname,
        )
        paths, _, _, _ = self._setup_paths(tmpdir)
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            with mock.patch(M_PATH + "load_userdata") as m_lud:
                m_lud.return_value = "ud"
                assert 1 == query.handle_args("anyname", args)
        assert expected_error in caplog.text

    def test_handle_args_error_on_missing_instance_data(self, caplog, tmpdir):
        """When instance_data file path does not exist, log an error."""
        absent_fn = tmpdir.join("absent")
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=absent_fn.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        assert 1 == query.handle_args("anyname", args)

        msg = "Missing instance-data file: %s" % absent_fn
        assert msg in caplog.text

    def test_handle_args_error_when_no_read_permission_instance_data(
        self, caplog, tmpdir
    ):
        """When instance_data file is unreadable, log an error."""
        noread_fn = tmpdir.join("unreadable")
        noread_fn.write("thou shall not pass")
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=noread_fn.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        with mock.patch(M_PATH + "util.load_binary_file") as m_load:
            m_load.side_effect = OSError(errno.EACCES, "Not allowed")
            assert 1 == query.handle_args("anyname", args)
        msg = "No read permission on '%s'. Try sudo" % noread_fn
        assert msg in caplog.text

    @pytest.mark.parametrize(
        "exception",
        [
            (OSError(errno.EACCES, "Not allowed"),),
            (OSError(errno.ENOENT, "Not allowed"),),
            (IOError,),
        ],
    )
    def test_handle_args_error_when_no_read_permission_init_cfg(
        self, exception, capsys
    ):
        """query.handle_status_args exists with 1 and no sys-output."""
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=None,
        )
        with mock.patch(
            M_PATH + "read_cfg_paths",
            side_effect=exception,
        ) as m_read_cfg_paths:
            query.handle_args("anyname", args)
        assert m_read_cfg_paths.call_count == 1
        out, err = capsys.readouterr()
        assert not out
        assert not err

    def test_handle_args_defaults_instance_data(self, caplog, tmpdir):
        """When no instance_data argument, default to configured run_dir."""
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=None,
        )
        paths, _, _, _ = self._setup_paths(tmpdir)
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            assert 1 == query.handle_args("anyname", args)
        json_file = paths.get_runpath("instance_data")
        msg = f"Missing instance-data file: {json_file}"
        assert msg in caplog.text

    def test_handle_args_root_fallsback_to_instance_data(self, caplog, tmpdir):
        """When no instance_data argument, root falls back to redacted json."""
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=None,
        )
        paths, _, _, _ = self._setup_paths(tmpdir)
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            with mock.patch("os.getuid") as m_getuid:
                m_getuid.return_value = 0
                assert 1 == query.handle_args("anyname", args)
        json_file = paths.get_runpath("instance_data")
        sensitive_file = paths.get_runpath("instance_data_sensitive")
        msg = (
            f"Missing root-readable {sensitive_file}. "
            f"Using redacted {json_file} instead."
        )
        assert msg in caplog.text

    @pytest.mark.parametrize(
        "ud_src,ud_expected,vd_src,vd_expected",
        (
            ("hi mom", "hi mom", "hi pops", "hi pops"),
            ("ud".encode("utf-8"), "ud", "vd".encode("utf-8"), "vd"),
            (_gzip_data(b"ud"), "ud", _gzip_data(b"vd"), "vd"),
            (_gzip_data("ud".encode("utf-8")), "ud", _gzip_data(b"vd"), "vd"),
        ),
    )
    def test_handle_args_root_processes_user_data(
        self, ud_src, ud_expected, vd_src, vd_expected, capsys, tmpdir
    ):
        """Support reading multiple user-data file content types"""
        paths, _, user_data, vendor_data = self._setup_paths(
            tmpdir, ud_val=ud_src, vd_val=vd_src
        )
        sensitive_file = Path(paths.get_runpath("instance_data_sensitive"))
        sensitive_file.write_text('{"my-var": "it worked"}')
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=user_data.strpath,
            vendor_data=vendor_data.strpath,
            varname=None,
        )
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            with mock.patch("os.getuid") as m_getuid:
                m_getuid.return_value = 0
                assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        cmd_output = json.loads(out)
        assert "it worked" == cmd_output["my-var"]
        if ud_expected == "ci-b64:":
            ud_expected = "ci-b64:{}".format(b64e(ud_src))
        if vd_expected == "ci-b64:":
            vd_expected = "ci-b64:{}".format(b64e(vd_src))
        assert ud_expected == cmd_output["userdata"]
        assert vd_expected == cmd_output["vendordata"]

    def test_handle_args_user_vendor_data_defaults_to_instance_link(
        self, capsys, tmpdir
    ):
        """When no instance_data argument, root uses sensitive json."""
        paths, _, _, _ = self._setup_paths(tmpdir)
        sensitive_file = Path(paths.get_runpath("instance_data_sensitive"))
        sensitive_file.write_text('{"my-var": "it worked"}')

        ud_path = os.path.join(paths.instance_link, "user-data.txt")
        write_file(ud_path, "instance_link_ud")
        vd_path = os.path.join(paths.instance_link, "vendor-data.txt")
        write_file(vd_path, "instance_link_vd")

        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=None,
            vendor_data=None,
            varname=None,
        )
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            with mock.patch("os.getuid", return_value=0):
                assert 0 == query.handle_args("anyname", args)
        expected = (
            '{\n "combined_cloud_config": null,\n "my-var": "it worked",\n '
            '"userdata": "instance_link_ud",\n '
            '"vendordata": "instance_link_vd"\n}\n'
        )
        out, _ = capsys.readouterr()
        assert expected == out

    def test_handle_args_root_uses_instance_sensitive_data(
        self, capsys, tmpdir
    ):
        """When no instance_data argument, root uses sensitive json."""
        paths, _, user_data, vendor_data = self._setup_paths(
            tmpdir, ud_val="ud", vd_val="vd"
        )
        sensitive_file = Path(paths.get_runpath("instance_data_sensitive"))
        sensitive_file.write_text('{"my-var": "it worked"}')
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=None,
            list_keys=False,
            user_data=user_data.strpath,
            vendor_data=vendor_data.strpath,
            varname=None,
        )
        with mock.patch(M_PATH + "read_cfg_paths") as m_paths:
            m_paths.return_value = paths
            with mock.patch("os.getuid") as m_getuid:
                m_getuid.return_value = 0
                assert 0 == query.handle_args("anyname", args)
        expected = (
            '{\n "combined_cloud_config": null,\n "my-var": "it worked",\n '
            '"userdata": "ud",\n "vendordata": "vd"\n}\n'
        )
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_dumps_all_instance_data(self, capsys, tmpdir):
        """When --all is specified query will dump all instance data vars."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write('{"my-var": "it worked"}')
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        expected = (
            '{\n "combined_cloud_config": "<redacted for non-root user> file:'
            '/run/cloud-init/combined-cloud-config.json",\n "my-var":'
            ' "it worked",\n "userdata": "<%s> file:ud",\n'
            ' "vendordata": "<%s> file:vd"\n}\n'
            % (REDACT_SENSITIVE_VALUE, REDACT_SENSITIVE_VALUE)
        )
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_returns_top_level_varname(self, capsys, tmpdir):
        """When the argument varname is passed, report its value."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write('{"my-var": "it worked"}')
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname="my_var",
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert "it worked\n" == out

    @pytest.mark.parametrize(
        "inst_data,varname,expected",
        (
            (
                '{"v1": {"key-2": "value-2"}, "my-var": "it worked"}',
                "v1.key_2",
                "value-2\n",
            ),
            # Assert no jinja underscore-delimited aliases are reported on CLI
            (
                '{"v1": {"something-hyphenated": {"no.underscores":"x",'
                ' "no-alias": "y"}}, "my-var": "it worked"}',
                "v1.something_hyphenated",
                '{\n "no-alias": "y",\n "no.underscores": "x"\n}\n',
            ),
        ),
    )
    def test_handle_args_returns_nested_varname(
        self, inst_data, varname, expected, capsys, tmpdir
    ):
        """If user_data file is a jinja template render instance-data vars."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(inst_data)
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=instance_data.strpath,
            user_data="ud",
            vendor_data="vd",
            list_keys=False,
            varname=varname,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_returns_standardized_vars_to_top_level_aliases(
        self, capsys, tmpdir
    ):
        """Any standardized vars under v# are promoted as top-level aliases."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1"}, "v2": {"v2_2": "val2.2"},'
            ' "top": "gun"}'
        )
        expected = dedent(
            """\
            {
             "combined_cloud_config": "<redacted for non-root user> %s",
             "top": "gun",
             "userdata": "<redacted for non-root user> file:ud",
             "v1": {
              "v1_1": "val1.1"
             },
             "v1_1": "val1.1",
             "v2": {
              "v2_2": "val2.2"
             },
             "v2_2": "val2.2",
             "vendordata": "<redacted for non-root user> file:vd"
            }
        """
            % "file:/run/cloud-init/combined-cloud-config.json"
        )
        args = self.Args(
            debug=False,
            dump_all=True,
            format=None,
            instance_data=instance_data.strpath,
            user_data="ud",
            vendor_data="vd",
            list_keys=False,
            varname=None,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_list_keys_sorts_top_level_keys_when_no_varname(
        self, capsys, tmpdir
    ):
        """Sort all top-level keys when only --list-keys provided."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1"}, "v2": {"v2_2": "val2.2"},'
            ' "top": "gun"}'
        )
        expected = (
            "combined_cloud_config\ntop\nuserdata\nv1\nv1_1\nv2\nv2_2\n"
            "vendordata\n"
        )
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=True,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_list_keys_sorts_nested_keys_when_varname(
        self, capsys, tmpdir
    ):
        """Sort all nested keys of varname object when --list-keys provided."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1", "v1_2": "val1.2"}, "v2":'
            ' {"v2_2": "val2.2"}, "top": "gun"}'
        )
        expected = "v1_1\nv1_2\n"
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=True,
            user_data="ud",
            vendor_data="vd",
            varname="v1",
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_list_keys_errors_when_varname_is_not_a_dict(
        self, caplog, tmpdir
    ):
        """Raise an error when --list-keys and varname specify a non-list."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1", "v1_2": "val1.2"}, "v2": '
            '{"v2_2": "val2.2"}, "top": "gun"}'
        )
        expected_error = "--list-keys provided but 'top' is not a dict"
        args = self.Args(
            debug=False,
            dump_all=False,
            format=None,
            instance_data=instance_data.strpath,
            list_keys=True,
            user_data="ud",
            vendor_data="vd",
            varname="top",
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 1 == query.handle_args("anyname", args)
        assert expected_error in caplog.text

    @pytest.mark.parametrize(
        "header_included",
        [True, False],
    )
    def test_handle_args_formats_jinja_successfully(
        self, caplog, tmpdir, capsys, header_included
    ):
        """Test that rendering a jinja template works as expected."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1", "v1_2": "val1.2"}, "v2": '
            '{"v2_2": "val2.2"}, "top": "gun"}'
        )
        header = "## template: jinja\n" if header_included else ""
        format = header + "v1_1: {{ v1.v1_1 }}"
        expected = header + "v1_1: val1.1\n"

        args = self.Args(
            debug=False,
            dump_all=False,
            format=format,
            instance_data=instance_data.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 0 == query.handle_args("anyname", args)
        out, _err = capsys.readouterr()
        assert expected == out

    def test_handle_args_invalid_jinja_exception(self, caplog, tmpdir, capsys):
        """Raise an error when a jinja syntax error is encountered."""
        instance_data = tmpdir.join("instance-data")
        instance_data.write(
            '{"v1": {"v1_1": "val1.1", "v1_2": "val1.2"}, "v2": '
            '{"v2_2": "val2.2"}, "top": "gun"}'
        )
        format = "v1_1: {{ v1.v1_1 } }"
        expected_error = (
            "Failed to render templated data. "
            + JinjaSyntaxParsingException.format_error_message(
                syntax_error="unexpected '}'",
                line_number=2,
                line_content="v1_1: {{ v1.v1_1 } }",
            )
        )
        args = self.Args(
            debug=False,
            dump_all=False,
            format=format,
            instance_data=instance_data.strpath,
            list_keys=False,
            user_data="ud",
            vendor_data="vd",
            varname=None,
        )
        with mock.patch("os.getuid") as m_getuid:
            m_getuid.return_value = 100
            assert 1 == query.handle_args("anyname", args)
        assert expected_error in caplog.text



# Copyright (C) 2016 Canonical Ltd.
#
# Author: Ryan Harper <ryan.harper@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""NTP: enable and configure ntp"""

import copy
import logging
import os
from textwrap import dedent

from cloudinit import subp, temp_utils, templater, type_utils, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.settings import PER_INSTANCE

LOG = logging.getLogger(__name__)

frequency = PER_INSTANCE
NTP_CONF = "/etc/ntp.conf"
NR_POOL_SERVERS = 4
distros = [
    "almalinux",
    "alpine",
    "centos",
    "cloudlinux",
    "cos",
    "debian",
    "eurolinux",
    "fedora",
    "freebsd",
    "mariner",
    "miraclelinux",
    "openbsd",
    "openeuler",
    "OpenCloudOS",
    "openmandriva",
    "opensuse",
    "opensuse-microos",
    "opensuse-tumbleweed",
    "opensuse-leap",
    "photon",
    "rhel",
    "rocky",
    "sle_hpc",
    "sle-micro",
    "sles",
    "TencentOS",
    "ubuntu",
    "virtuozzo",
]

NTP_CLIENT_CONFIG = {
    "chrony": {
        "check_exe": "chronyd",
        "confpath": "/etc/chrony.conf",
        "packages": ["chrony"],
        "service_name": "chrony",
        "template_name": "chrony.conf.{distro}",
        "template": None,
    },
    "ntp": {
        "check_exe": "ntpd",
        "confpath": NTP_CONF,
        "packages": ["ntp"],
        "service_name": "ntp",
        "template_name": "ntp.conf.{distro}",
        "template": None,
    },
    "ntpdate": {
        "check_exe": "ntpdate",
        "confpath": NTP_CONF,
        "packages": ["ntpdate"],
        "service_name": "ntpdate",
        "template_name": "ntp.conf.{distro}",
        "template": None,
    },
    "openntpd": {
        "check_exe": "ntpd",
        "confpath": "/etc/ntpd.conf",
        "packages": [],
        "service_name": "ntpd",
        "template_name": "ntpd.conf.{distro}",
        "template": None,
    },
    "systemd-timesyncd": {
        "check_exe": "/lib/systemd/systemd-timesyncd",
        "confpath": "/etc/systemd/timesyncd.conf.d/cloud-init.conf",
        "packages": [],
        "service_name": "systemd-timesyncd",
        "template_name": "timesyncd.conf",
        "template": None,
    },
}

# This is Distro-specific configuration overrides of the base config
DISTRO_CLIENT_CONFIG = {
    "alpine": {
        "chrony": {
            "confpath": "/etc/chrony/chrony.conf",
            "service_name": "chronyd",
        },
        "ntp": {
            "confpath": "/etc/ntp.conf",
            "packages": [],
            "service_name": "ntpd",
        },
    },
    "centos": {
        "ntp": {
            "service_name": "ntpd",
        },
        "chrony": {
            "service_name": "chronyd",
        },
    },
    "cos": {
        "chrony": {
            "service_name": "chronyd",
            "confpath": "/etc/chrony/chrony.conf",
        },
    },
    "debian": {
        "chrony": {
            "confpath": "/etc/chrony/chrony.conf",
        },
    },
    "freebsd": {
        "ntp": {
            "confpath": "/etc/ntp.conf",
            "service_name": "ntpd",
            "template_name": "ntp.conf.{distro}",
        },
        "chrony": {
            "confpath": "/usr/local/etc/chrony.conf",
            "packages": ["chrony"],
            "service_name": "chronyd",
            "template_name": "chrony.conf.{distro}",
        },
        "openntpd": {
            "check_exe": "/usr/local/sbin/ntpd",
            "confpath": "/usr/local/etc/ntp.conf",
            "packages": ["openntpd"],
            "service_name": "openntpd",
            "template_name": "ntpd.conf.openbsd",
        },
    },
    "mariner": {
        "chrony": {
            "service_name": "chronyd",
        },
        "systemd-timesyncd": {
            "check_exe": "/usr/lib/systemd/systemd-timesyncd",
            "confpath": "/etc/systemd/timesyncd.conf",
        },
    },
    "openbsd": {
        "openntpd": {},
    },
    "openmandriva": {
        "chrony": {
            "service_name": "chronyd",
        },
        "ntp": {
            "confpath": "/etc/ntp.conf",
            "service_name": "ntpd",
        },
        "systemd-timesyncd": {
            "check_exe": "/lib/systemd/systemd-timesyncd",
        },
    },
    "opensuse": {
        "chrony": {
            "service_name": "chronyd",
        },
        "ntp": {
            "confpath": "/etc/ntp.conf",
            "service_name": "ntpd",
        },
        "systemd-timesyncd": {
            "check_exe": "/usr/lib/systemd/systemd-timesyncd",
        },
    },
    "photon": {
        "chrony": {
            "service_name": "chronyd",
        },
        "ntp": {"service_name": "ntpd", "confpath": "/etc/ntp.conf"},
        "systemd-timesyncd": {
            "check_exe": "/usr/lib/systemd/systemd-timesyncd",
            "confpath": "/etc/systemd/timesyncd.conf",
        },
    },
    "rhel": {
        "ntp": {
            "service_name": "ntpd",
        },
        "chrony": {
            "service_name": "chronyd",
        },
    },
    "sles": {
        "chrony": {
            "service_name": "chronyd",
        },
        "ntp": {
            "confpath": "/etc/ntp.conf",
            "service_name": "ntpd",
        },
        "systemd-timesyncd": {
            "check_exe": "/usr/lib/systemd/systemd-timesyncd",
        },
    },
    "ubuntu": {
        "chrony": {
            "confpath": "/etc/chrony/chrony.conf",
        },
    },
}

for distro in ("opensuse-microos", "opensuse-tumbleweed", "opensuse-leap"):
    DISTRO_CLIENT_CONFIG[distro] = DISTRO_CLIENT_CONFIG["opensuse"]

for distro in ("sle_hpc", "sle-micro"):
    DISTRO_CLIENT_CONFIG[distro] = DISTRO_CLIENT_CONFIG["sles"]

# The schema definition for each cloud-config module is a strict contract for
# describing supported configuration parameters for each cloud-config section.
# It allows cloud-config to validate and alert users to invalid or ignored
# configuration options before actually attempting to deploy with said
# configuration.

meta: MetaSchema = {
    "id": "cc_ntp",
    "name": "NTP",
    "title": "enable and configure ntp",
    "description": dedent(
        """\
        Handle ntp configuration. If ntp is not installed on the system and
        ntp configuration is specified, ntp will be installed. If there is a
        default ntp config file in the image or one is present in the
        distro's ntp package, it will be copied to a file with ``.dist``
        appended to the filename before any changes are made. A list of ntp
        pools and ntp servers can be provided under the ``ntp`` config key.
        If no ntp ``servers`` or ``pools`` are provided, 4 pools will be used
        in the format ``{0-3}.{distro}.pool.ntp.org``."""
    ),
    "distros": distros,
    "examples": [
        dedent(
            """\
        # Override ntp with chrony configuration on Ubuntu
        ntp:
          enabled: true
          ntp_client: chrony  # Uses cloud-init default chrony configuration
        """
        ),
        dedent(
            """\
        # Provide a custom ntp client configuration
        ntp:
          enabled: true
          ntp_client: myntpclient
          config:
             confpath: /etc/myntpclient/myntpclient.conf
             check_exe: myntpclientd
             packages:
               - myntpclient
             service_name: myntpclient
             template: |
                 ## template:jinja
                 # My NTP Client config
                 {% if pools -%}# pools{% endif %}
                 {% for pool in pools -%}
                 pool {{pool}} iburst
                 {% endfor %}
                 {%- if servers %}# servers
                 {% endif %}
                 {% for server in servers -%}
                 server {{server}} iburst
                 {% endfor %}
                 {% if peers -%}# peers{% endif %}
                 {% for peer in peers -%}
                 peer {{peer}}
                 {% endfor %}
                 {% if allow -%}# allow{% endif %}
                 {% for cidr in allow -%}
                 allow {{cidr}}
                 {% endfor %}
          pools: [0.int.pool.ntp.org, 1.int.pool.ntp.org, ntp.myorg.org]
          servers:
            - ntp.server.local
            - ntp.ubuntu.com
            - 192.168.23.2
          allow:
            - 192.168.23.0/32
          peers:
            - km001
            - km002"""
        ),
    ],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": ["ntp"],
}
__doc__ = get_meta_doc(meta)


REQUIRED_NTP_CONFIG_KEYS = frozenset(
    ["check_exe", "confpath", "packages", "service_name"]
)


def distro_ntp_client_configs(distro):
    """Construct a distro-specific ntp client config dictionary by merging
       distro specific changes into base config.

    @param distro: String providing the distro class name.
    @returns: Dict of distro configurations for ntp clients.
    """
    dcfg = DISTRO_CLIENT_CONFIG
    cfg = copy.copy(NTP_CLIENT_CONFIG)
    if distro in dcfg:
        cfg = util.mergemanydict([cfg, dcfg[distro]], reverse=True)
    return cfg


def select_ntp_client(ntp_client, distro):
    """Determine which ntp client is to be used, consulting the distro
       for its preference.

    @param ntp_client: String name of the ntp client to use.
    @param distro: Distro class instance.
    @returns: Dict of the selected ntp client or {} if none selected.
    """

    # construct distro-specific ntp_client_config dict
    distro_cfg = distro_ntp_client_configs(distro.name)

    # user specified client, return its config
    if ntp_client and ntp_client != "auto":
        LOG.debug(
            'Selected NTP client "%s" via user-data configuration', ntp_client
        )
        return distro_cfg.get(ntp_client, {})

    # default to auto if unset in distro
    distro_ntp_client = distro.get_option("ntp_client", "auto")

    clientcfg = {}
    if distro_ntp_client == "auto":
        for client in distro.preferred_ntp_clients:
            cfg = distro_cfg.get(client)
            if subp.which(cfg.get("check_exe")):
                LOG.debug(
                    'Selected NTP client "%s", already installed', client
                )
                clientcfg = cfg
                break

        if not clientcfg:
            client = distro.preferred_ntp_clients[0]
            LOG.debug(
                'Selected distro preferred NTP client "%s", not yet installed',
                client,
            )
            clientcfg = distro_cfg.get(client)
    else:
        LOG.debug(
            'Selected NTP client "%s" via distro system config',
            distro_ntp_client,
        )
        clientcfg = distro_cfg.get(distro_ntp_client, {})

    return clientcfg


def install_ntp_client(install_func, packages=None, check_exe="ntpd"):
    """Install ntp client package if not already installed.

    @param install_func: function.  This parameter is invoked with the contents
    of the packages parameter.
    @param packages: list.  This parameter defaults to ['ntp'].
    @param check_exe: string.  The name of a binary that indicates the package
    the specified package is already installed.
    """
    if subp.which(check_exe):
        return
    if packages is None:
        packages = ["ntp"]

    install_func(packages)


def rename_ntp_conf(confpath=None):
    """Rename any existing ntp client config file

    @param confpath: string. Specify a path to an existing ntp client
    configuration file.
    """
    if os.path.exists(confpath):
        util.rename(confpath, confpath + ".dist")


def generate_server_names(distro):
    """Generate a list of server names to populate an ntp client configuration
    file.

    @param distro: string.  Specify the distro name
    @returns: list: A list of strings representing ntp servers for this distro.
    """
    names = []
    pool_distro = distro

    if distro == "sles":
        # For legal reasons x.pool.sles.ntp.org does not exist,
        # use the opensuse pool
        pool_distro = "opensuse"
    elif distro == "alpine" or distro == "eurolinux":
        # Alpine-specific pool (i.e. x.alpine.pool.ntp.org) does not exist
        # so use general x.pool.ntp.org instead. The same applies to EuroLinux
        pool_distro = ""

    for x in range(NR_POOL_SERVERS):
        names.append(
            ".".join(
                [n for n in [str(x)] + [pool_distro] + ["pool.ntp.org"] if n]
            )
        )

    return names


def write_ntp_config_template(
    distro_name,
    service_name=None,
    servers=None,
    pools=None,
    allow=None,
    peers=None,
    path=None,
    template_fn=None,
    template=None,
):
    """Render a ntp client configuration for the specified client.

    @param distro_name: string.  The distro class name.
    @param service_name: string. The name of the NTP client service.
    @param servers: A list of strings specifying ntp servers. Defaults to empty
    list.
    @param pools: A list of strings specifying ntp pools. Defaults to empty
    list.
    @param allow: A list of strings specifying a network/CIDR. Defaults to
    empty list.
    @param peers: A list nodes that should peer with each other. Defaults to
    empty list.
    @param path: A string to specify where to write the rendered template.
    @param template_fn: A string to specify the template source file.
    @param template: A string specifying the contents of the template. This
    content will be written to a temporary file before being used to render
    the configuration file.

    @raises: ValueError when path is None.
    @raises: ValueError when template_fn is None and template is None.
    """
    if not servers:
        servers = []
    if not pools:
        pools = []
    if not allow:
        allow = []
    if not peers:
        peers = []

    if len(servers) == 0 and len(pools) == 0 and distro_name == "cos":
        return
    if (
        len(servers) == 0
        and distro_name == "alpine"
        and service_name == "ntpd"
    ):
        # Alpine's Busybox ntpd only understands "servers" configuration
        # and not "pool" configuration.
        servers = generate_server_names(distro_name)
        LOG.debug("Adding distro default ntp servers: %s", ",".join(servers))
    elif len(servers) == 0 and len(pools) == 0:
        pools = generate_server_names(distro_name)
        LOG.debug(
            "Adding distro default ntp pool servers: %s", ",".join(pools)
        )

    if not path:
        raise ValueError("Invalid value for path parameter")

    if not template_fn and not template:
        raise ValueError("Not template_fn or template provided")

    params = {
        "servers": servers,
        "pools": pools,
        "allow": allow,
        "peers": peers,
    }
    if template:
        tfile = temp_utils.mkstemp(prefix="template_name-", suffix=".tmpl")
        template_fn = tfile[1]  # filepath is second item in tuple
        util.write_file(template_fn, content=template)

    templater.render_to_file(template_fn, path, params)
    # clean up temporary template
    if template:
        util.del_file(template_fn)


def supplemental_schema_validation(ntp_config):
    """Validate user-provided ntp:config option values.

    This function supplements flexible jsonschema validation with specific
    value checks to aid in triage of invalid user-provided configuration.

    @param ntp_config: Dictionary of configuration value under 'ntp'.

    @raises: ValueError describing invalid values provided.
    """
    errors = []
    missing = REQUIRED_NTP_CONFIG_KEYS.difference(set(ntp_config.keys()))
    if missing:
        keys = ", ".join(sorted(missing))
        errors.append(
            "Missing required ntp:config keys: {keys}".format(keys=keys)
        )
    elif not any(
        [ntp_config.get("template"), ntp_config.get("template_name")]
    ):
        errors.append(
            "Either ntp:config:template or ntp:config:template_name values"
            " are required"
        )
    for key, value in sorted(ntp_config.items()):
        keypath = "ntp:config:" + key
        if key == "confpath":
            if not all([value, isinstance(value, str)]):
                errors.append(
                    "Expected a config file path {keypath}."
                    " Found ({value})".format(keypath=keypath, value=value)
                )
        elif key == "packages":
            if not isinstance(value, list):
                errors.append(
                    "Expected a list of required package names for {keypath}."
                    " Found ({value})".format(keypath=keypath, value=value)
                )
        elif key in ("template", "template_name"):
            if value is None:  # Either template or template_name can be none
                continue
            if not isinstance(value, str):
                errors.append(
                    "Expected a string type for {keypath}."
                    " Found ({value})".format(keypath=keypath, value=value)
                )
        elif not isinstance(value, str):
            errors.append(
                "Expected a string type for {keypath}. Found ({value})".format(
                    keypath=keypath, value=value
                )
            )

    if errors:
        raise ValueError(
            r"Invalid ntp configuration:\n{errors}".format(
                errors="\n".join(errors)
            )
        )


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    """Enable and configure ntp."""
    if "ntp" not in cfg:
        LOG.debug(
            "Skipping module named %s, not present or disabled by cfg", name
        )
        return
    ntp_cfg = cfg["ntp"]
    if ntp_cfg is None:
        ntp_cfg = {}  # Allow empty config which will install the package

    # TODO drop this when validate_cloudconfig_schema is strict=True
    if not isinstance(ntp_cfg, (dict)):
        raise RuntimeError(
            "'ntp' key existed in config, but not a dictionary type,"
            " is a {_type} instead".format(_type=type_utils.obj_name(ntp_cfg))
        )

    # Allow users to explicitly enable/disable
    enabled = ntp_cfg.get("enabled", True)
    if util.is_false(enabled):
        LOG.debug("Skipping module named %s, disabled by cfg", name)
        return

    # Select which client is going to be used and get the configuration
    ntp_client_config = select_ntp_client(
        ntp_cfg.get("ntp_client"), cloud.distro
    )
    # Allow user ntp config to override distro configurations
    ntp_client_config = util.mergemanydict(
        [ntp_client_config, ntp_cfg.get("config", {})], reverse=True
    )

    supplemental_schema_validation(ntp_client_config)
    rename_ntp_conf(confpath=ntp_client_config.get("confpath"))

    template_fn = None
    if not ntp_client_config.get("template"):
        template_name = ntp_client_config.get("template_name").replace(
            "{distro}", cloud.distro.name
        )
        template_fn = cloud.get_template_filename(template_name)
        if not template_fn:
            msg = (
                "No template found, not rendering %s"
                % ntp_client_config.get("template_name")
            )
            raise RuntimeError(msg)

    LOG.debug("service_name: %s", ntp_client_config.get("service_name"))
    LOG.debug("servers: %s", ntp_cfg.get("servers", []))
    LOG.debug("pools: %s", ntp_cfg.get("pools", []))
    LOG.debug("allow: %s", ntp_cfg.get("allow", []))
    LOG.debug("peers: %s", ntp_cfg.get("peers", []))
    write_ntp_config_template(
        cloud.distro.name,
        service_name=ntp_client_config.get("service_name"),
        servers=ntp_cfg.get("servers", []),
        pools=ntp_cfg.get("pools", []),
        allow=ntp_cfg.get("allow", []),
        peers=ntp_cfg.get("peers", []),
        path=ntp_client_config.get("confpath"),
        template_fn=template_fn,
        template=ntp_client_config.get("template"),
    )

    install_ntp_client(
        cloud.distro.install_packages,
        packages=ntp_client_config["packages"],
        check_exe=ntp_client_config["check_exe"],
    )
    if util.is_BSD():
        if ntp_client_config.get("service_name") != "ntpd":
            try:
                cloud.distro.manage_service("stop", "ntpd")
            except subp.ProcessExecutionError:
                LOG.warning("Failed to stop base ntpd service")
            try:
                cloud.distro.manage_service("disable", "ntpd")
            except subp.ProcessExecutionError:
                LOG.warning("Failed to disable base ntpd service")

        try:
            cloud.distro.manage_service(
                "enable", ntp_client_config.get("service_name")
            )
        except subp.ProcessExecutionError as e:
            LOG.exception("Failed to enable ntp service: %s", e)
            raise
    try:
        cloud.distro.manage_service(
            "reload", ntp_client_config.get("service_name")
        )
    except subp.ProcessExecutionError as e:
        LOG.exception("Failed to reload/start ntp service: %s", e)
        raise



# This file is part of cloud-init. See LICENSE file for license information.
import json
import logging
import re
import sys
from collections import namedtuple

import jsonschema
import pytest

from cloudinit import subp
from cloudinit.config.cc_ubuntu_pro import (
    _attach,
    _auto_attach,
    _should_auto_attach,
    configure_pro,
    handle,
    maybe_install_ua_tools,
    set_pro_config,
    validate_schema_features,
)
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from cloudinit.util import Version
from tests.unittests.helpers import does_not_raise, mock, skipUnlessJsonSchema
from tests.unittests.util import get_cloud

# Module path used in mocks
MPATH = "cloudinit.config.cc_ubuntu_pro"


class FakeUserFacingError(Exception):
    def __init__(self, msg: str):
        self.msg = msg


class FakeAlreadyAttachedError(FakeUserFacingError):
    pass


class FakeAlreadyAttachedOnPROError(FakeUserFacingError):
    pass


@pytest.fixture
def fake_uaclient(mocker):
    """Mocks `uaclient` module"""

    mocker.patch.dict("sys.modules")
    m_uaclient = mock.Mock()

    sys.modules["uaclient"] = m_uaclient

    # Exceptions
    _exceptions = namedtuple(
        "exceptions",
        [
            "UserFacingError",
            "AlreadyAttachedError",
        ],
    )(
        FakeUserFacingError,
        FakeAlreadyAttachedError,
    )
    sys.modules["uaclient.api.exceptions"] = _exceptions


@pytest.mark.usefixtures("fake_uaclient")
@mock.patch(f"{MPATH}.subp.subp")
class TestConfigurePro:
    def test_configure_pro_attach_error(self, m_subp):
        """Errors from pro attach command are raised."""
        m_subp.side_effect = subp.ProcessExecutionError(
            "Invalid token SomeToken"
        )
        match = (
            "Failure attaching Ubuntu Pro:\nUnexpected error while"
            " running command.\nCommand: -\nExit code: -\nReason: -\n"
            "Stdout: Invalid token REDACTED\nStderr: -"
        )
        with pytest.raises(RuntimeError, match=match):
            configure_pro(token="SomeToken")

    @pytest.mark.parametrize(
        "kwargs, call_args_list, log_record_tuples",
        [
            # When token is provided, attach to pro using the token.
            pytest.param(
                {"token": "SomeToken"},
                [
                    mock.call(
                        ["pro", "attach", "SomeToken"],
                        logstring=["pro", "attach", "REDACTED"],
                        rcs={0, 2},
                    )
                ],
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Attaching to Ubuntu Pro. pro attach REDACTED",
                    )
                ],
                id="with_token",
            ),
            # When services is an empty list, do not auto-enable attach.
            pytest.param(
                {"token": "SomeToken", "enable": []},
                [
                    mock.call(
                        ["pro", "attach", "SomeToken"],
                        logstring=["pro", "attach", "REDACTED"],
                        rcs={0, 2},
                    )
                ],
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Attaching to Ubuntu Pro. pro attach REDACTED",
                    )
                ],
                id="with_empty_services",
            ),
            # When services a list, only enable specific services.
            pytest.param(
                {"token": "SomeToken", "enable": ["fips"]},
                [
                    mock.call(
                        ["pro", "attach", "--no-auto-enable", "SomeToken"],
                        logstring=[
                            "pro",
                            "attach",
                            "--no-auto-enable",
                            "REDACTED",
                        ],
                        rcs={0, 2},
                    ),
                    mock.call(
                        [
                            "pro",
                            "enable",
                            "--assume-yes",
                            "--format",
                            "json",
                            "fips",
                        ],
                        capture=True,
                        rcs={0, 1},
                    ),
                ],
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Attaching to Ubuntu Pro. pro attach"
                        " --no-auto-enable REDACTED",
                    )
                ],
                id="with_specific_services",
            ),
            # When services a string, treat as singleton list and warn
            pytest.param(
                {"token": "SomeToken", "enable": "fips"},
                [
                    mock.call(
                        ["pro", "attach", "--no-auto-enable", "SomeToken"],
                        logstring=[
                            "pro",
                            "attach",
                            "--no-auto-enable",
                            "REDACTED",
                        ],
                        rcs={0, 2},
                    ),
                    mock.call(
                        [
                            "pro",
                            "enable",
                            "--assume-yes",
                            "--format",
                            "json",
                            "fips",
                        ],
                        capture=True,
                        rcs={0, 1},
                    ),
                ],
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Attaching to Ubuntu Pro. pro attach"
                        " --no-auto-enable REDACTED",
                    ),
                    (
                        MPATH,
                        logging.WARNING,
                        "ubuntu_pro: enable should be a list, not a "
                        "string; treating as a single enable",
                    ),
                ],
                id="with_string_services",
            ),
            # When services not string or list, warn but still attach
            pytest.param(
                {"token": "SomeToken", "enable": {"deffo": "wont work"}},
                [
                    mock.call(
                        ["pro", "attach", "SomeToken"],
                        logstring=["pro", "attach", "REDACTED"],
                        rcs={0, 2},
                    )
                ],
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Attaching to Ubuntu Pro. pro attach REDACTED",
                    ),
                    (
                        MPATH,
                        logging.WARNING,
                        "ubuntu_pro: enable should be a list, not a"
                        " dict; skipping enabling services",
                    ),
                ],
                id="with_weird_services",
            ),
        ],
    )
    @mock.patch(f"{MPATH}.maybe_install_ua_tools", mock.MagicMock())
    def test_configure_pro_attach(
        self, m_subp, kwargs, call_args_list, log_record_tuples, caplog
    ):
        m_subp.return_value = subp.SubpResult(json.dumps({"errors": []}), "")
        configure_pro(**kwargs)
        assert call_args_list == m_subp.call_args_list
        for record_tuple in log_record_tuples:
            assert record_tuple in caplog.record_tuples

    def test_configure_pro_already_attached(self, m_subp, caplog):
        """pro is already attached to an subscription"""
        m_subp.rcs = 2
        configure_pro(token="SomeToken")
        assert m_subp.call_args_list == [
            mock.call(
                ["pro", "attach", "SomeToken"],
                logstring=["pro", "attach", "REDACTED"],
                rcs={0, 2},
            )
        ]
        assert (
            MPATH,
            logging.DEBUG,
            "Attaching to Ubuntu Pro. pro attach REDACTED",
        ) in caplog.record_tuples

    def test_configure_pro_attach_on_service_enabled(
        self, m_subp, caplog, fake_uaclient
    ):
        """retry enabling an already enabled service"""

        def fake_subp(cmd, capture=None, rcs=None, logstring=None):
            fail_cmds = [
                "pro",
                "enable",
                "--assume-yes",
                "--format",
                "json",
                "livepatch",
            ]
            if cmd == fail_cmds and capture:
                response = {
                    "errors": [
                        {
                            "message": "Does not matter",
                            "message_code": "service-already-enabled",
                            "service": cmd[-1],
                            "type": "service",
                        }
                    ]
                }
                return subp.SubpResult(json.dumps(response), "")

        m_subp.side_effect = fake_subp

        configure_pro(token="SomeToken", enable=["livepatch"])
        assert m_subp.call_args_list == [
            mock.call(
                ["pro", "attach", "--no-auto-enable", "SomeToken"],
                logstring=["pro", "attach", "--no-auto-enable", "REDACTED"],
                rcs={0, 2},
            ),
            mock.call(
                [
                    "pro",
                    "enable",
                    "--assume-yes",
                    "--format",
                    "json",
                    "livepatch",
                ],
                capture=True,
                rcs={0, 1},
            ),
        ]
        assert (
            MPATH,
            logging.DEBUG,
            "Service `livepatch` already enabled.",
        ) in caplog.record_tuples

    def test_configure_pro_attach_on_service_error(self, m_subp, caplog):
        """all services should be enabled and then any failures raised"""

        def fake_subp(cmd, capture=None, rcs=None, logstring=None):
            fail_cmd = [
                "pro",
                "enable",
                "--assume-yes",
                "--format",
                "json",
            ]
            if cmd[: len(fail_cmd)] == fail_cmd and capture:
                response = {
                    "errors": [
                        {
                            "message": f"Invalid {svc} credentials",
                            "message_code": "some-code",
                            "service": svc,
                            "type": "service",
                        }
                        for svc in ["esm", "cc"]
                    ]
                    + [
                        {
                            "message": "Cannot enable unknown service 'asdf'",
                            "message_code": "invalid-service-or-failure",
                            "service": None,
                            "type": "system",
                        }
                    ]
                }
                return subp.SubpResult(json.dumps(response), "")
            return subp.SubpResult(json.dumps({"errors": []}), "")

        m_subp.side_effect = fake_subp

        with pytest.raises(
            RuntimeError,
            match=re.escape("Failure enabling Ubuntu Pro service(s): esm, cc"),
        ):
            configure_pro(
                token="SomeToken", enable=["esm", "cc", "fips", "asdf"]
            )
        assert m_subp.call_args_list == [
            mock.call(
                ["pro", "attach", "--no-auto-enable", "SomeToken"],
                logstring=["pro", "attach", "--no-auto-enable", "REDACTED"],
                rcs={0, 2},
            ),
            mock.call(
                [
                    "pro",
                    "enable",
                    "--assume-yes",
                    "--format",
                    "json",
                    "esm",
                    "cc",
                    "fips",
                    "asdf",
                ],
                capture=True,
                rcs={0, 1},
            ),
        ]
        assert (
            MPATH,
            logging.WARNING,
            "Failure enabling `esm`: Invalid esm credentials",
        ) in caplog.record_tuples
        assert (
            MPATH,
            logging.WARNING,
            "Failure enabling `cc`: Invalid cc credentials",
        ) in caplog.record_tuples
        assert (
            MPATH,
            logging.WARNING,
            "Failure of type `system`: Cannot enable unknown service 'asdf'",
        ) in caplog.record_tuples
        assert 'Failure enabling "fips"' not in caplog.text

    def test_pro_enable_unexpected_error_codes(self, m_subp):
        def fake_subp(cmd, capture=None, **kwargs):
            if cmd[:2] == ["pro", "enable"] and capture:
                raise subp.ProcessExecutionError(exit_code=255)
            return subp.SubpResult(json.dumps({"errors": []}), "")

        m_subp.side_effect = fake_subp

        with pytest.raises(
            RuntimeError,
            match=re.escape("Error while enabling service(s): esm"),
        ):
            configure_pro(token="SomeToken", enable=["esm"])

    def test_pro_enable_non_json_response(self, m_subp):
        def fake_subp(cmd, capture=None, **kwargs):
            if cmd[:2] == ["pro", "enable"] and capture:
                return subp.SubpResult("I dream to be a Json", "")
            return subp.SubpResult(json.dumps({"errors": []}), "")

        m_subp.side_effect = fake_subp

        with pytest.raises(
            RuntimeError,
            match=re.escape("Pro response was not json: I dream to be a Json"),
        ):
            configure_pro(token="SomeToken", enable=["esm"])


JSONSCHEMA_SKIP_REASON = (
    "deprecation unraised as jsonschema ver can't merge $defs and inline keys"
)


class TestUbuntuProSchema:
    @pytest.mark.parametrize(
        "config, expectation, skip_reason",
        [
            pytest.param({"ubuntu_pro": {}}, does_not_raise(), ""),
            pytest.param(
                {"ubuntu_advantage": {}},
                pytest.raises(
                    SchemaValidationError,
                    match=re.escape(
                        "ubuntu_advantage:  Deprecated in version 24.1."
                        " Use ``ubuntu_pro`` instead"
                    ),
                ),
                # If __version__ no longer exists on jsonschema, that means
                # we're using a high enough version of jsonschema to not need
                # to skip this test.
                JSONSCHEMA_SKIP_REASON
                if Version.from_str(getattr(jsonschema, "__version__", "999"))
                < Version(4)
                else "",
                id="deprecation_of_ubuntu_advantage_skip_old_json",
            ),
            # Strict keys
            pytest.param(
                {"ubuntu_pro": {"token": "win", "invalidkey": ""}},
                pytest.raises(
                    SchemaValidationError,
                    match=re.escape(
                        "ubuntu_pro: Additional properties are not"
                        " allowed ('invalidkey"
                    ),
                ),
                "",
                id="additional_properties",
            ),
            pytest.param(
                {"ubuntu_pro": {"features": {"disable_auto_attach": True}}},
                does_not_raise(),
                "",
                id="disable_auto_attach",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "features": {"disable_auto_attach": False},
                        "enable": ["fips"],
                        "enable_beta": ["realtime-kernel"],
                        "token": "<token>",
                    }
                },
                does_not_raise(),
                "",
                id="pro_custom_services",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "enable_beta": ["realtime-kernel"],
                        "token": "<token>",
                    }
                },
                does_not_raise(),
                "",
                id="non_pro_beta_services",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "features": {"asdf": False},
                        "enable": ["fips"],
                        "enable_beta": ["realtime-kernel"],
                        "token": "<token>",
                    }
                },
                pytest.raises(
                    SchemaValidationError,
                    match=re.escape(
                        "ubuntu_pro.features: Additional properties are"
                        " not allowed ('asdf'"
                    ),
                ),
                "",
                id="pro_additional_features",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "token": "<token>",
                        "config": {
                            "http_proxy": "http://some-proxy:8088",
                            "https_proxy": "https://some-proxy:8088",
                            "global_apt_https_proxy": "https://some-global-apt-proxy:8088/",  # noqa: E501
                            "global_apt_http_proxy": "http://some-global-apt-proxy:8088/",  # noqa: E501
                            "ua_apt_http_proxy": "http://10.0.10.10:3128",
                            "ua_apt_https_proxy": "https://10.0.10.10:3128",
                        },
                    }
                },
                does_not_raise(),
                "",
                id="pro_config_valid_set",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "token": "<token>",
                        "config": {
                            "http_proxy": None,
                            "https_proxy": None,
                            "global_apt_https_proxy": None,
                            "global_apt_http_proxy": None,
                            "ua_apt_http_proxy": None,
                            "ua_apt_https_proxy": None,
                        },
                    }
                },
                does_not_raise(),
                "",
                id="pro_config_valid_unset",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "token": "<token>",
                        "config": ["http_proxy=http://some-proxy:8088"],
                    }
                },
                pytest.raises(
                    SchemaValidationError,
                    match=re.escape(
                        "errors: ubuntu_pro.config:"
                        " ['http_proxy=http://some-proxy:8088']"
                    ),
                ),
                "",
                id="pro_config_invalid_type",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "token": "<token>",
                        "config": {
                            "http_proxy": 8888,
                            "https_proxy": ["http://some-proxy:8088"],
                        },
                    }
                },
                pytest.raises(
                    SchemaValidationError,
                    match=re.escape(
                        "errors: ubuntu_pro.config.http_proxy: 8888"
                        " is not of type 'string', 'null',"
                        " ubuntu_pro.config.https_proxy:"
                        " ['http://some-proxy:8088']"
                    ),
                ),
                "",
                id="pro_config_invalid_proxy_type",
            ),
            pytest.param(
                {
                    "ubuntu_pro": {
                        "enable": ["fips"],
                        "token": "<token>",
                        "config": {
                            "http_proxy": "http://some-proxy:8088",
                            "hola": "adios",
                        },
                    }
                },
                does_not_raise(),
                "",
                id="pro_config_unknown_props_allowed",
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, expectation, skip_reason, caplog):
        if skip_reason:
            pytest.skip(skip_reason)
        with expectation:
            validate_cloudconfig_schema(config, get_schema(), strict=True)

    @pytest.mark.parametrize(
        "ua_section, expectation, log_msgs",
        [
            ({}, does_not_raise(), None),
            ({"features": {}}, does_not_raise(), None),
            (
                {"features": {"disable_auto_attach": True}},
                does_not_raise(),
                None,
            ),
            (
                {"features": {"disable_auto_attach": False}},
                does_not_raise(),
                None,
            ),
            (
                {"features": [0, 1]},
                pytest.raises(
                    RuntimeError,
                    match=(
                        "'ubuntu_pro.features' should be a dict, not a list"
                    ),
                ),
                ["'ubuntu_pro.features' should be a dict, not a list\n"],
            ),
            (
                {"features": {"disable_auto_attach": [0, 1]}},
                pytest.raises(
                    RuntimeError,
                    match=(
                        "'ubuntu_pro.features.disable_auto_attach'"
                        " should be a bool, not a list"
                    ),
                ),
                [
                    "'ubuntu_pro.features.disable_auto_attach' should be"
                    " a bool, not a list\n"
                ],
            ),
        ],
    )
    def test_validate_schema_features(
        self, ua_section, expectation, log_msgs, caplog
    ):
        with expectation:
            validate_schema_features(ua_section)
        if log_msgs is not None:
            for log_msg in log_msgs:
                assert log_msg in caplog.text
        else:
            assert not caplog.text


class TestHandle:

    cloud = get_cloud()

    @pytest.mark.parametrize(
        [
            "cfg",
            "cloud",
            "log_record_tuples",
            "maybe_install_call_args_list",
            "set_pro_config_call_args_list",
            "configure_pro_call_args_list",
        ],
        [
            # When no ua-related configuration is provided, nothing happens.
            pytest.param(
                {},
                None,
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Skipping module named nomatter, no 'ubuntu_pro'"
                        " configuration found",
                    )
                ],
                [],
                [],
                [],
                id="no_config",
            ),
            # If ubuntu_pro is provided, try installing ua-tools package.
            pytest.param(
                {"ubuntu_pro": {"token": "valid"}},
                cloud,
                [],
                [mock.call(cloud)],
                [mock.call(None)],
                None,
                id="tries_to_install_ubuntu_advantage_tools",
            ),
            # If ubuntu_pro config provided, configure it.
            pytest.param(
                {
                    "ubuntu_pro": {
                        "token": "valid",
                        "config": {"http_proxy": "http://proxy.org"},
                    }
                },
                cloud,
                [],
                None,
                [mock.call({"http_proxy": "http://proxy.org"})],
                None,
                id="set_pro_config",
            ),
            # All ubuntu_pro config keys are passed to configure_pro.
            pytest.param(
                {"ubuntu_pro": {"token": "token", "enable": ["esm"]}},
                cloud,
                [],
                [mock.call(cloud)],
                [mock.call(None)],
                [mock.call(token="token", enable=["esm"])],
                id="passes_credentials_and_services_to_configure_pro",
            ),
            # Warning when ubuntu-advantage key is present with new config
            pytest.param(
                {"ubuntu-advantage": {"token": "token", "enable": ["esm"]}},
                None,
                [
                    (
                        MPATH,
                        logging.WARNING,
                        "Deprecated configuration key(s) provided:"
                        ' ubuntu-advantage. Expected "ubuntu_pro"; will'
                        " attempt to continue.",
                    ),
                ],
                None,
                [mock.call(None)],
                [mock.call(token="token", enable=["esm"])],
                id="warns_on_deprecated_ubuntu_pro_key_w_config",
            ),
            # Warning with beta services during attach
            pytest.param(
                {
                    "ubuntu_pro": {
                        "token": "token",
                        "enable": ["esm"],
                        "enable_beta": ["realtime-kernel"],
                    }
                },
                None,
                [
                    (
                        MPATH,
                        logging.DEBUG,
                        "Ignoring `ubuntu_pro.enable_beta` services in"
                        " Pro attach: realtime-kernel",
                    )
                ],
                None,
                [mock.call(None)],
                [mock.call(token="token", enable=["esm"])],
                id="warns_on_enable_beta_in_attach",
            ),
            # ubuntu_pro should be preferred over ubuntu-advantage
            pytest.param(
                {
                    "ubuntu-advantage": {"token": "nope", "enable": ["wrong"]},
                    "ubuntu_pro": {"token": "token", "enable": ["esm"]},
                },
                None,
                [
                    (
                        MPATH,
                        logging.WARNING,
                        "Deprecated configuration key(s) provided:"
                        ' ubuntu-advantage. Expected "ubuntu_pro"; will'
                        " attempt to continue.",
                    ),
                    (
                        MPATH,
                        logging.WARNING,
                        "Ignoring deprecated key ubuntu-advantage and"
                        " preferring ubuntu_pro config",
                    ),
                ],
                None,
                [mock.call(None)],
                [mock.call(token="token", enable=["esm"])],
                id="prefers_new_style_config",
            ),
        ],
    )
    @mock.patch(f"{MPATH}._should_auto_attach", return_value=False)
    @mock.patch(f"{MPATH}._auto_attach")
    @mock.patch(f"{MPATH}.configure_pro")
    @mock.patch(f"{MPATH}.set_pro_config")
    @mock.patch(f"{MPATH}.maybe_install_ua_tools")
    def test_handle_attach(
        self,
        m_maybe_install_ua_tools,
        m_set_pro_config,
        m_configure_pro,
        m_auto_attach,
        m_should_auto_attach,
        cfg,
        cloud,
        log_record_tuples,
        maybe_install_call_args_list,
        set_pro_config_call_args_list,
        configure_pro_call_args_list,
        caplog,
    ):
        """Non-Pro schemas and instance."""
        handle("nomatter", cfg=cfg, cloud=cloud, args=None)
        for record_tuple in log_record_tuples:
            assert record_tuple in caplog.record_tuples
        if maybe_install_call_args_list is not None:
            assert (
                maybe_install_call_args_list
                == m_maybe_install_ua_tools.call_args_list
            )
        if set_pro_config_call_args_list is not None:
            assert (
                set_pro_config_call_args_list
                == m_set_pro_config.call_args_list
            )
        if configure_pro_call_args_list is not None:
            assert (
                configure_pro_call_args_list == m_configure_pro.call_args_list
            )
        assert [] == m_auto_attach.call_args_list

    @pytest.mark.parametrize(
        [
            "cfg",
            "cloud",
            "log_record_tuples",
            "auto_attach_side_effect",
            "should_auto_attach",
            "auto_attach_call_args_list",
            "attach_call_args_list",
            "expectation",
        ],
        [
            # When auto_attach successes, no call to configure_pro.
            pytest.param(
                {"ubuntu_pro": {"features": {"disable_auto_attach": False}}},
                cloud,
                [],
                None,  # auto_attach successes
                True,  # Pro instance
                [
                    mock.call({"features": {"disable_auto_attach": False}})
                ],  # auto_attach_call_args_list
                [],  # attach_call_args_list
                does_not_raise(),
                id="auto_attach_success",
            ),
            # When auto_attach fails in a Pro instance, no call to
            # configure_pro.
            pytest.param(
                {"ubuntu_pro": {"features": {"disable_auto_attach": False}}},
                cloud,
                [],
                RuntimeError("Auto attach error"),
                True,  # Pro instance
                [
                    mock.call({"features": {"disable_auto_attach": False}})
                ],  # auto_attach_call_args_list
                [],  # attach_call_args_list
                pytest.raises(RuntimeError, match="Auto attach error"),
                id="auto_attach_error",
            ),
            # In a non-Pro instance with token, fallback to normal attach.
            pytest.param(
                {
                    "ubuntu_pro": {
                        "features": {"disable_auto_attach": False},
                        "token": "token",
                    }
                },
                cloud,
                [],
                None,
                False,  # non-Pro instance
                [],  # auto_attach_call_args_list
                [
                    mock.call(
                        {
                            "features": {"disable_auto_attach": False},
                            "token": "token",
                        },
                    )
                ],  # attach_call_args_list
                does_not_raise(),
                id="not_pro_with_token",
            ),
            # In a non-Pro instance with enable, fallback to normal attach.
            pytest.param(
                {"ubuntu_pro": {"enable": ["esm"]}},
                cloud,
                [],
                None,
                False,  # non-Pro instance
                [],  # auto_attach_call_args_list
                [
                    mock.call(
                        {
                            "enable": ["esm"],
                        },
                    )
                ],  # attach_call_args_list
                does_not_raise(),
                id="not_pro_with_enable",
            ),
        ],
    )
    @mock.patch(f"{MPATH}._should_auto_attach")
    @mock.patch(f"{MPATH}._auto_attach")
    @mock.patch(f"{MPATH}._attach")
    def test_handle_auto_attach_vs_attach(
        self,
        m_attach,
        m_auto_attach,
        m_should_auto_attach,
        cfg,
        cloud,
        log_record_tuples,
        auto_attach_side_effect,
        should_auto_attach,
        auto_attach_call_args_list,
        attach_call_args_list,
        expectation,
        caplog,
    ):
        m_should_auto_attach.return_value = should_auto_attach
        if auto_attach_side_effect is not None:
            m_auto_attach.side_effect = auto_attach_side_effect

        with expectation:
            handle("nomatter", cfg=cfg, cloud=cloud, args=None)

        for record_tuple in log_record_tuples:
            assert record_tuple in caplog.record_tuples
        if attach_call_args_list is not None:
            assert attach_call_args_list == m_attach.call_args_list
        else:
            assert [] == m_attach.call_args_list
        assert auto_attach_call_args_list == m_auto_attach.call_args_list

    @pytest.mark.parametrize("is_pro", [False, True])
    @pytest.mark.parametrize(
        "cfg",
        [
            (
                {
                    "ubuntu_pro": {
                        "features": {"disable_auto_attach": False},
                    }
                }
            ),
            (
                {
                    "ubuntu_pro": {
                        "features": {"disable_auto_attach": True},
                    }
                }
            ),
        ],
    )
    @mock.patch(f"{MPATH}._should_auto_attach")
    @mock.patch(f"{MPATH}._auto_attach")
    @mock.patch(f"{MPATH}._attach")
    def test_no_fallback_attach(
        self,
        m_attach,
        m_auto_attach,
        m_should_auto_attach,
        cfg,
        is_pro,
    ):
        """Checks that attach is not called in the case where we want only to
        enable or disable pro auto-attach.
        """
        m_should_auto_attach.return_value = is_pro
        handle("nomatter", cfg=cfg, cloud=self.cloud, args=None)
        assert not m_attach.call_args_list

    @pytest.mark.parametrize(
        "cfg, handle_kwargs, match",
        [
            pytest.param(
                {"ubuntu-advantage": {"commands": "nogo"}},
                dict(cloud=None, args=None),
                (
                    'Deprecated configuration "ubuntu-advantage: commands" '
                    'provided. Expected "token"'
                ),
                id="key_dashed",
            ),
            pytest.param(
                {"ubuntu_pro": {"commands": "nogo"}},
                dict(cloud=None, args=None),
                (
                    'Deprecated configuration "ubuntu-advantage: commands" '
                    'provided. Expected "token"'
                ),
                id="key_underscore",
            ),
        ],
    )
    @mock.patch("%s.configure_pro" % MPATH)
    def test_handle_error_on_deprecated_commands_key_dashed(
        self, m_configure_pro, cfg, handle_kwargs, match
    ):
        with pytest.raises(RuntimeError, match=match):
            handle("nomatter", cfg=cfg, **handle_kwargs)
        assert 0 == m_configure_pro.call_count

    @pytest.mark.parametrize(
        "cfg, match",
        [
            pytest.param(
                {"ubuntu_pro": [0, 1]},
                "'ubuntu_pro' should be a dict, not a list",
                id="on_non_dict_config",
            ),
            pytest.param(
                {"ubuntu_pro": {"features": [0, 1]}},
                "'ubuntu_pro.features' should be a dict, not a list",
                id="on_non_dict_ua_section",
            ),
        ],
    )
    def test_handle_errors(self, cfg, match):
        with pytest.raises(RuntimeError, match=match):
            handle(
                "nomatter",
                cfg=cfg,
                cloud=self.cloud,
                args=None,
            )

    @mock.patch(f"{MPATH}.subp.subp")
    def test_pro_config_error_invalid_url(self, m_subp, caplog):
        """Errors from pro config command are raised."""
        cfg = {
            "ubuntu_pro": {
                "token": "SomeToken",
                "config": {"http_proxy": "not-a-valid-url"},
            }
        }
        m_subp.side_effect = subp.ProcessExecutionError(
            'Failure enabling "http_proxy"'
        )
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Invalid ubuntu_pro configuration:\nExpected URL scheme"
                " http/https for ua:config:http_proxy"
            ),
        ):
            handle(
                "nomatter",
                cfg=cfg,
                cloud=self.cloud,
                args=None,
            )
        assert not caplog.text

    @mock.patch(f"{MPATH}._should_auto_attach", return_value=False)
    @mock.patch(f"{MPATH}.subp.subp")
    def test_fallback_to_attach_no_token(
        self, m_subp, m_should_auto_attach, caplog
    ):
        cfg = {"ubuntu_pro": {"enable": ["esm"]}}
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "`ubuntu_pro.token` required in non-Pro Ubuntu instances."
            ),
        ):
            handle(
                "nomatter",
                cfg=cfg,
                cloud=self.cloud,
                args=None,
            )
        assert [] == m_subp.call_args_list
        assert (
            "`ubuntu_pro.token` required in non-Pro Ubuntu instances.\n"
        ) in caplog.text


class TestShouldAutoAttach:
    def test_should_auto_attach_error(self, caplog, fake_uaclient):
        m_should_auto_attach = mock.Mock()
        m_should_auto_attach.should_auto_attach.side_effect = (
            FakeUserFacingError("Some error")  # noqa: E501
        )
        sys.modules[
            "uaclient.api.u.pro.attach.auto.should_auto_attach.v1"
        ] = m_should_auto_attach
        assert not _should_auto_attach({})
        assert "Error during `should_auto_attach`: Some error" in caplog.text
        assert (
            "Unable to determine if this is an Ubuntu Pro instance."
            " Fallback to normal Pro attach." in caplog.text
        )

    @pytest.mark.parametrize(
        "ua_section, expected_result",
        [
            ({}, None),
            ({"features": {"disable_auto_attach": False}}, None),
            # The user explicitly disables auto-attach, therefore we do not do
            # it:
            ({"features": {"disable_auto_attach": True}}, False),
        ],
    )
    def test_happy_path(
        self, ua_section, expected_result, caplog, fake_uaclient
    ):
        m_should_auto_attach = mock.Mock()
        sys.modules[
            "uaclient.api.u.pro.attach.auto.should_auto_attach.v1"
        ] = m_should_auto_attach
        should_auto_attach_value = object()
        m_should_auto_attach.should_auto_attach.return_value.should_auto_attach = (  # noqa: E501
            should_auto_attach_value
        )
        if expected_result is None:  # Pro API does respond
            assert should_auto_attach_value == _should_auto_attach(ua_section)
            assert (
                "Checking if the instance can be attached to Ubuntu Pro took"
                in caplog.text
            )
        else:  # cloud-init does respond
            assert expected_result == _should_auto_attach(ua_section)
            assert not caplog.text


class TestAutoAttach:

    ua_section: dict = {}

    def test_full_auto_attach_error(self, caplog, mocker, fake_uaclient):
        mocker.patch.dict("sys.modules")
        sys.modules["uaclient.config"] = mock.Mock()
        m_full_auto_attach = mock.Mock()
        m_full_auto_attach.full_auto_attach.side_effect = FakeUserFacingError(
            "Some error"
        )
        sys.modules[
            "uaclient.api.u.pro.attach.auto.full_auto_attach.v1"
        ] = m_full_auto_attach
        expected_msg = "Error during `full_auto_attach`: Some error"
        with pytest.raises(RuntimeError, match=re.escape(expected_msg)):
            _auto_attach(self.ua_section)
        assert expected_msg in caplog.text

    def test_happy_path(self, caplog, mocker, fake_uaclient):
        mocker.patch.dict("sys.modules")
        sys.modules["uaclient.config"] = mock.Mock()
        sys.modules[
            "uaclient.api.u.pro.attach.auto.full_auto_attach.v1"
        ] = mock.Mock()
        _auto_attach(self.ua_section)
        assert "Attaching to Ubuntu Pro took" in caplog.text


class TestAttach:
    @mock.patch(f"{MPATH}.configure_pro")
    def test_attach_without_token_raises_error(self, m_configure_pro):
        with pytest.raises(
            RuntimeError,
            match=("`ubuntu_pro.token` required in non-Pro Ubuntu instances."),
        ):
            _attach({"enable": ["esm"]})
        assert [] == m_configure_pro.call_args_list


@mock.patch(f"{MPATH}.subp.which")
class TestMaybeInstallUATools:
    @pytest.mark.parametrize(
        [
            "which_return",
            "update_side_effect",
            "install_side_effect",
            "expectation",
            "log_msg",
        ],
        [
            # Do nothing if ubuntu-advantage-tools already exists.
            pytest.param(
                "/usr/bin/ua",  # already installed
                RuntimeError("Some apt error"),
                None,
                does_not_raise(),  # No RuntimeError
                None,
                id="noop_when_ua_tools_present",
            ),
            # logs and raises apt update errors
            pytest.param(
                None,
                RuntimeError("Some apt error"),
                None,
                pytest.raises(RuntimeError, match="Some apt error"),
                "Package update failed\nTraceback",
                id="raises_update_errors",
            ),
            # logs and raises package install errors
            pytest.param(
                None,
                None,
                RuntimeError("Some install error"),
                pytest.raises(RuntimeError, match="Some install error"),
                "Failed to install ubuntu-advantage-tools\n",
                id="raises_install_errors",
            ),
        ],
    )
    def test_maybe_install_ua_tools(
        self,
        m_which,
        which_return,
        update_side_effect,
        install_side_effect,
        expectation,
        log_msg,
        caplog,
    ):
        m_which.return_value = which_return
        cloud = mock.MagicMock()
        if install_side_effect is None:
            cloud.distro.update_package_sources.side_effect = (
                update_side_effect
            )
        else:
            cloud.distro.update_package_sources.return_value = None
            cloud.distro.install_packages.side_effect = install_side_effect
        with expectation:
            maybe_install_ua_tools(cloud=cloud)
        if log_msg is not None:
            assert log_msg in caplog.text

    def test_maybe_install_ua_tools_happy_path(self, m_which):
        """maybe_install_ua_tools installs ubuntu-advantage-tools."""
        m_which.return_value = None
        cloud = mock.MagicMock()  # No errors raised
        maybe_install_ua_tools(cloud=cloud)
        assert [
            mock.call()
        ] == cloud.distro.update_package_sources.call_args_list
        assert [
            mock.call(["ubuntu-advantage-tools"])
        ] == cloud.distro.install_packages.call_args_list


@mock.patch(f"{MPATH}.subp.subp")
class TestSetProConfig:
    def test_valid_config(self, m_subp, caplog):
        pro_config = {
            "http_proxy": "http://some-proxy:8088",
            "https_proxy": "https://user:pass@some-proxy:8088",
            "global_apt_https_proxy": "https://some-global-apt-proxy:8088/",
            "global_apt_http_proxy": "http://some-global-apt-proxy:8088/",
            "ua_apt_http_proxy": "http://10.0.10.10:3128",
            "ua_apt_https_proxy": "https://10.0.10.10:3128",
        }
        set_pro_config(pro_config)
        for ua_arg, redacted_arg in [
            (
                "http_proxy=http://some-proxy:8088",
                "http_proxy=REDACTED",
            ),
            (
                "https_proxy=https://user:pass@some-proxy:8088",
                "https_proxy=REDACTED",
            ),
            (
                "global_apt_https_proxy=https://some-global-apt-proxy:8088/",
                "global_apt_https_proxy=REDACTED",
            ),
            (
                "global_apt_http_proxy=http://some-global-apt-proxy:8088/",
                "global_apt_http_proxy=REDACTED",
            ),
            (
                "ua_apt_http_proxy=http://10.0.10.10:3128",
                "ua_apt_http_proxy=REDACTED",
            ),
            (
                "ua_apt_https_proxy=https://10.0.10.10:3128",
                "ua_apt_https_proxy=REDACTED",
            ),
        ]:
            assert (
                mock.call(
                    ["pro", "config", "set", ua_arg],
                    logstring=["pro", "config", "set", redacted_arg],
                )
                in m_subp.call_args_list
            )
            assert f"Enabling Pro config {redacted_arg}\n" in caplog.text
            assert ua_arg not in caplog.text

        assert 6 == m_subp.call_count

    def test_pro_config_unset(self, m_subp, caplog):
        pro_config = {
            "https_proxy": "https://user:pass@some-proxy:8088",
            "http_proxy": None,
        }
        set_pro_config(pro_config)
        for call in [
            mock.call(["pro", "config", "unset", "http_proxy"]),
            mock.call(
                [
                    "pro",
                    "config",
                    "set",
                    "https_proxy=https://user:pass@some-proxy:8088",
                ],
                logstring=["pro", "config", "set", "https_proxy=REDACTED"],
            ),
        ]:
            assert call in m_subp.call_args_list
        assert 2 == m_subp.call_count
        assert "Enabling Pro config https_proxy=REDACTED\n" in caplog.text
        assert "https://user:pass@some-proxy:8088" not in caplog.text
        assert "Disabling Pro config for http_proxy\n" in caplog.text

    def test_pro_config_error_non_string_values(self, m_subp, caplog):
        """ValueError raised for any values expected as string type."""
        pro_config = {
            "global_apt_http_proxy": "noscheme",
            "http_proxy": ["no-proxy"],
            "https_proxy": 3.14,
        }
        match = re.escape(
            "Invalid ubuntu_pro configuration:\n"
            "Expected URL scheme http/https for"
            " ua:config:global_apt_http_proxy\n"
            "Expected a URL for ua:config:http_proxy\n"
            "Expected a URL for ua:config:https_proxy"
        )
        with pytest.raises(ValueError, match=match):
            set_pro_config(pro_config)
        assert 0 == m_subp.call_count
        assert not caplog.text

    def test_pro_config_unknown_prop(self, m_subp, caplog):
        """On unknown config props, a log is issued and the prop is set."""
        pro_config = {"asdf": "qwer"}
        set_pro_config(pro_config)
        assert [
            mock.call(
                ["pro", "config", "set", "asdf=qwer"],
                logstring=["pro", "config", "set", "asdf=REDACTED"],
            )
        ] == m_subp.call_args_list
        assert "qwer" not in caplog.text
        assert (
            "Not validating unknown ubuntu_pro.config.asdf property\n"
            in caplog.text
        )

    def test_pro_config_wrong_type(self, m_subp, caplog):
        pro_config = ["asdf", "qwer"]
        with pytest.raises(
            RuntimeError,
            match=(
                "ubuntu_pro: config should be a dict, not"
                " a list; skipping enabling config parameters"
            ),
        ):
            set_pro_config(pro_config)
        assert 0 == m_subp.call_count
        assert not caplog.text

    def test_set_pro_config_error(self, m_subp, caplog):
        pro_config = {
            "https_proxy": "https://user:pass@some-proxy:8088",
        }
        # Simulate Pro error
        m_subp.side_effect = subp.ProcessExecutionError(
            "Invalid proxy: https://user:pass@some-proxy:8088"
        )
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Failure enabling/disabling Ubuntu Pro config(s):"
                ' "https_proxy"'
            ),
        ):
            set_pro_config(pro_config)
        assert 1 == m_subp.call_count
        assert "https://user:pass@some-proxy:8088" not in caplog.text
        assert "Enabling Pro config https_proxy=REDACTED\n" in caplog.text
        assert 'Failure enabling/disabling "https_proxy":\n' in caplog.text

    def test_unset_pro_config_error(self, m_subp, caplog):
        pro_config = {"https_proxy": None}
        # Simulate Pro error
        m_subp.side_effect = subp.ProcessExecutionError(
            "Error unsetting https_proxy"
        )
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Failure enabling/disabling Ubuntu Pro config(s): "
                '"https_proxy"'
            ),
        ):
            set_pro_config(pro_config)
        assert 1 == m_subp.call_count
        assert "https://user:pass@some-proxy:8088" not in caplog.text
        assert "Disabling Pro config for https_proxy\n" in caplog.text
        assert 'Failure enabling/disabling "https_proxy":\n' in caplog.text



# This file is part of cloud-init. See LICENSE file for license information.

import contextlib
import errno
import logging
import os
import shutil
import tempfile

from cloudinit import util

LOG = logging.getLogger(__name__)
_TMPDIR = None
_ROOT_TMPDIR = "/run/cloud-init/tmp"
_EXE_ROOT_TMPDIR = "/var/tmp/cloud-init"


def get_tmp_ancestor(odir=None, needs_exe: bool = False):
    if odir is not None:
        return odir
    if needs_exe:
        return _EXE_ROOT_TMPDIR
    if _TMPDIR:
        return _TMPDIR
    if os.getuid() == 0:
        return _ROOT_TMPDIR
    return os.environ.get("TMPDIR", "/tmp")


def _tempfile_dir_arg(odir=None, needs_exe: bool = False):
    """Return the proper 'dir' argument for tempfile functions.

    When root, cloud-init will use /run/cloud-init/tmp to avoid
    any cleaning that a distro boot might do on /tmp (such as
    systemd-tmpfiles-clean).

    If the caller of this function (mkdtemp or mkstemp) was provided
    with a 'dir' argument, then that is respected.

    @param odir: original 'dir' arg to 'mkdtemp' or other.
    @param needs_exe: Boolean specifying whether or not exe permissions are
        needed for tempdir. This is needed because /run is mounted noexec.
    """
    tdir = get_tmp_ancestor(odir, needs_exe)
    if not os.path.isdir(tdir):
        os.makedirs(tdir)
        os.chmod(tdir, 0o1777)

    if needs_exe:
        if util.has_mount_opt(tdir, "noexec"):
            LOG.warning(
                "Requested temporal dir with exe permission `%s` is"
                " mounted as noexec",
                tdir,
            )

    if odir is None and not needs_exe:
        global _TMPDIR
        _TMPDIR = tdir

    return tdir


def ExtendedTemporaryFile(**kwargs):
    kwargs["dir"] = _tempfile_dir_arg(
        kwargs.pop("dir", None), kwargs.pop("needs_exe", False)
    )
    fh = tempfile.NamedTemporaryFile(**kwargs)
    # Replace its unlink with a quiet version
    # that does not raise errors when the
    # file to unlink has been unlinked elsewhere..

    def _unlink_if_exists(path):
        try:
            os.unlink(path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise e

    fh.unlink = _unlink_if_exists

    # Add a new method that will unlink
    # right 'now' but still lets the exit
    # method attempt to remove it (which will
    # not throw due to our del file being quiet
    # about files that are not there)
    def unlink_now():
        fh.unlink(fh.name)

    setattr(fh, "unlink_now", unlink_now)
    return fh


@contextlib.contextmanager
def tempdir(rmtree_ignore_errors=False, **kwargs):
    # This seems like it was only added in python 3.2
    # Make it since its useful...
    # See: http://bugs.python.org/file12970/tempdir.patch
    tdir = mkdtemp(**kwargs)
    try:
        yield tdir
    finally:
        shutil.rmtree(tdir, ignore_errors=rmtree_ignore_errors)


def mkdtemp(dir=None, needs_exe: bool = False, **kwargs):
    dir = _tempfile_dir_arg(dir, needs_exe)
    return tempfile.mkdtemp(dir=dir, **kwargs)


def mkstemp(dir=None, needs_exe: bool = False, **kwargs):
    dir = _tempfile_dir_arg(dir, needs_exe)
    return tempfile.mkstemp(dir=dir, **kwargs)



import pytest

from cloudinit import sources
from cloudinit.sources import DataSourceOpenStack as ds
from tests.unittests.helpers import mock

openstack_ds_name = ds.DataSourceOpenStack.dsname.lower()


@pytest.mark.parametrize(
    "m_cmdline, expected_ds",
    (
        # test ci.ds=
        ("aosiejfoij ci.ds=OpenStack ", openstack_ds_name),
        ("ci.ds=OpenStack", openstack_ds_name),
        ("aosiejfoij ci.ds=OpenStack blah", openstack_ds_name),
        ("aosiejfoij ci.ds=OpenStack faljskebflk", openstack_ds_name),
        ("ci.ds=OpenStack;", openstack_ds_name),
        ("ci.ds=openstack;", openstack_ds_name),
        ("notci.ds=somecloud ci.ds=openstack", openstack_ds_name),
        # test ci.datasource=
        ("aosiejfoij ci.datasource=OpenStack ", openstack_ds_name),
        ("ci.datasource=OpenStack", openstack_ds_name),
        ("aosiejfoij ci.datasource=OpenStack blah", openstack_ds_name),
        ("aosiejfoij ci.datasource=OpenStack faljskebflk", openstack_ds_name),
        ("ci.datasource=OpenStack;", openstack_ds_name),
        ("ci.datasource=openstack;", openstack_ds_name),
        ("notci.datasource=0 ci.datasource=nocloud", "nocloud"),
        # weird whitespace
        ("ci.datasource=OpenStack\n", openstack_ds_name),
        ("ci.datasource=OpenStack\t", openstack_ds_name),
        ("ci.datasource=OpenStack\r", openstack_ds_name),
        ("ci.datasource=OpenStack\v", openstack_ds_name),
        ("ci.ds=OpenStack\n", openstack_ds_name),
        ("ci.ds=OpenStack\t", openstack_ds_name),
        ("ci.ds=OpenStack\r", openstack_ds_name),
        ("ci.ds=OpenStack\v", openstack_ds_name),
        ("ci.ds=nocloud-net\v", "nocloud-net"),
        ("ci.datasource=nocloud\v", "nocloud"),
        # test ds=
        ("ds=nocloud-net", "nocloud-net"),
        ("foo ds=nocloud-net bar", "nocloud-net"),
        ("bonding.max_bonds=0", ""),
        ("foo bonding.max_bonds=0 ds=nocloud-net bar", "nocloud-net"),
    ),
)
def test_ds_detect_kernel_commandline(m_cmdline, expected_ds):
    """check commandline match"""
    with mock.patch(
        "cloudinit.util.get_cmdline",
        return_value=m_cmdline,
    ):
        assert (
            expected_ds == sources.parse_cmdline().lower()
        ), f"could not parse [{m_cmdline}]"



try:
    from hypothesis_jsonschema import from_schema

    HAS_HYPOTHESIS_JSONSCHEMA = True
except ImportError:
    HAS_HYPOTHESIS_JSONSCHEMA = False

    def from_schema(*_, **__):  # type: ignore
        pass


__all__ = ["from_schema", "HAS_HYPOTHESIS_JSONSCHEMA"]



"""Integration tests for LXD bridge creation.

(This is ported from
``tests/cloud_tests/testcases/modules/lxd_bridge.yaml``.)
"""
import warnings

import pytest
import yaml

from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import CURRENT_RELEASE, FOCAL
from tests.integration_tests.util import verify_clean_log

BRIDGE_USER_DATA = """\
#cloud-config
lxd:
  init:
    storage_backend: btrfs
  bridge:
    mode: new
    name: lxdbr0
    ipv4_address: 10.100.100.1
    ipv4_netmask: 24
    ipv4_dhcp_first: 10.100.100.100
    ipv4_dhcp_last: 10.100.100.200
    ipv4_nat: true
    domain: lxd
    mtu: 9000
"""

STORAGE_USER_DATA = """\
#cloud-config
lxd:
  init:
    storage_backend: {}
"""

PRESEED_USER_DATA = """\
#cloud-config
lxd:
  preseed: |
    config: {}
    networks:
    - config:
        ipv4.address: auto
        ipv6.address: auto
      description: ""
      name: lxdbr0
      type: ""
    storage_pools:
    - config:
        source: /var/snap/lxd/common/lxd/storage-pools/default
      description: ""
      name: default
      driver: dir
    profiles:
    - config: {}
      description: ""
      devices:
        eth0:
          name: eth0
          nictype: bridged
          parent: lxdbr0
          type: nic
        root:
          path: /
          pool: default
          type: disk
      name: default
    cluster: null
"""


STORAGE_PRESEED_USER_DATA = """\
#cloud-config
lxd:
  preseed: |
    networks:
    - config:
        ipv4.address: 10.42.42.1/24
        ipv4.nat: "true"
        ipv6.address: fd42:4242:4242:4242::1/64
        ipv6.nat: "true"
      description: ""
      name: lxdbr0
      type: bridge
      project: default
    storage_pools:
    - config:
        size: 5GiB
        source: /var/snap/lxd/common/lxd/disks/default.img
      description: ""
      name: default
      driver: {driver}
    profiles:
    - config: {{ }}
      description: Default LXD profile
      devices:
        eth0:
          {nictype}
          {parent}
          {network}
          type: nic
        root:
          path: /
          pool: default
          type: disk
      name: default
    - config:
        user.vendor-data: |
          #cloud-config
          write_files:
          - path: /var/lib/cloud/scripts/per-once/setup-lxc.sh
            encoding: b64
            permissions: '0755'
            owner: root:root
            content: |
              IyEvYmluL2Jhc2gKZWNobyBZRVAgPj4gL3Zhci9sb2cvY2xvdWQtaW5pdC5sb2cK
      devices:
        config:
          source: cloud-init:config
          type: disk
        eth0:
          name: eth0
          network: lxdbr0
          type: nic
        root:
          path: /
          pool: default
          type: disk
      description: Pycloudlib LXD profile for bionic VMs
      name: bionic-vm-lxc-setup
    projects:
    - config:
        features.images: "true"
        features.networks: "true"
        features.profiles: "true"
        features.storage.volumes: "true"
      description: Default LXD project
      name: default
    - config:
        features.images: "false"
        features.networks: "true"
        features.profiles: "false"
        features.storage.volumes: "true"
      description: Limited project
      name: limited
"""


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot run LXD"
)
@pytest.mark.user_data(BRIDGE_USER_DATA)
class TestLxdBridge:
    @pytest.mark.parametrize("binary_name", ["lxc", "lxd"])
    def test_binaries_installed(self, class_client, binary_name):
        """Check that the expected LXD binaries are installed"""
        assert class_client.execute(["which", binary_name]).ok

    def test_bridge(self, class_client):
        """Check that the given bridge is configured"""
        cloud_init_log = class_client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(cloud_init_log)

        # The bridge should exist
        assert class_client.execute("ip addr show lxdbr0").ok

        raw_network_config = class_client.execute("lxc network show lxdbr0")
        network_config = yaml.safe_load(raw_network_config)
        assert "10.100.100.1/24" == network_config["config"]["ipv4.address"]


def validate_storage(validate_client, pkg_name, command):
    log = validate_client.read_from_file("/var/log/cloud-init.log")
    verify_clean_log(log, ignore_deprecations=False)
    return log


def validate_preseed_profiles(client, preseed_cfg):
    for src_profile in preseed_cfg["profiles"]:
        profile = yaml.safe_load(
            client.execute(f"lxc profile show {src_profile['name']}")
        )
        assert src_profile["config"] == profile["config"]


def validate_preseed_storage_pools(client, preseed_cfg):
    for src_storage in preseed_cfg["storage_pools"]:
        storage_pool = yaml.safe_load(
            client.execute(f"lxc storage show {src_storage['name']}")
        )
        if "volatile.initial_source" in storage_pool["config"]:
            assert storage_pool["config"]["source"] == storage_pool[
                "config"
            ].pop("volatile.initial_source")
        if storage_pool["driver"] == "zfs":
            "default" == storage_pool["config"].pop("zfs.pool_name")
        assert storage_pool["config"] == src_storage["config"]
        assert storage_pool["driver"] == src_storage["driver"]


def validate_preseed_projects(client: IntegrationInstance, preseed_cfg):
    # Support for projects by lxd init --preseed was added in lxd 4.12
    # https://github.com/canonical/lxd/releases/tag/lxd-4.12
    if CURRENT_RELEASE.series in ("bionic", "focal"):
        return
    for src_project in preseed_cfg.get("projects", []):
        proj_name = src_project["name"]
        proj_result = client.execute(f"lxc project show {proj_name}")
        assert (
            proj_result.ok
        ), f"Error showing `{proj_name}` project info: {proj_result.stderr}"
        project = yaml.safe_load(
            client.execute(f"lxc project show {src_project['name']}")
        )
        project.pop("used_by", None)

        # `features.storage.buckets` was introduced in lxd 5.5. More info:
        # https://github.com/canonical/lxd/releases/tag/lxd-5.5
        if "features.storage.buckets" in project["config"]:
            assert "true" == project["config"].pop("features.storage.buckets")
        # `features.networks.zones` was introduced in lxd 5.9. More info:
        # https://github.com/canonical/lxd/releases/tag/lxd-5.9
        if "features.networks.zones" in project["config"]:
            assert "true" == project["config"].pop("features.networks.zones")
        assert project == src_project


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot manipulate storage"
)
@pytest.mark.user_data(STORAGE_USER_DATA.format("btrfs"))
def test_storage_btrfs(client):
    validate_storage(client, "btrfs-progs", "mkfs.btrfs")


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot manipulate LXD"
)
@pytest.mark.skipif(
    CURRENT_RELEASE < FOCAL, reason="tested on Focal and later"
)
def test_storage_preseed_btrfs(setup_image, session_cloud: IntegrationCloud):
    # TODO: If test is marked as not bionic, why is there a bionic section?
    if CURRENT_RELEASE.series in ("bionic",):
        nictype = "nictype: bridged"
        parent = "parent: lxdbr0"
        network = ""
    else:
        nictype = ""
        parent = ""
        network = "network: lxdbr0"
    user_data = STORAGE_PRESEED_USER_DATA.format(
        driver="btrfs", nictype=nictype, parent=parent, network=network
    )
    with session_cloud.launch(user_data=user_data) as client:
        validate_storage(client, "btrfs-progs", "mkfs.btrfs")
        src_cfg = yaml.safe_load(user_data)
        preseed_cfg = yaml.safe_load(src_cfg["lxd"]["preseed"])
        validate_preseed_profiles(client, preseed_cfg)
        validate_preseed_storage_pools(client, preseed_cfg)
        validate_preseed_projects(client, preseed_cfg)


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot manipulate LVM"
)
@pytest.mark.user_data(STORAGE_USER_DATA.format("lvm"))
def test_storage_lvm(client):
    log = client.read_from_file("/var/log/cloud-init.log")

    # Note to self
    if (
        "doesn't use thinpool by default on Ubuntu due to LP" not in log
        and "-kvm" not in client.execute("uname -r")
    ):
        warnings.warn("LP 1982780 has been fixed, update to allow thinpools")
    validate_storage(client, "lvm2", "lvcreate")


@pytest.mark.user_data(PRESEED_USER_DATA)
def test_basic_preseed(client):
    preseed_cfg = yaml.safe_load(PRESEED_USER_DATA)["lxd"]["preseed"]
    preseed_cfg = yaml.safe_load(preseed_cfg)
    cloud_init_log = client.read_from_file("/var/log/cloud-init.log")
    verify_clean_log(cloud_init_log)
    validate_preseed_profiles(client, preseed_cfg)
    validate_preseed_storage_pools(client, preseed_cfg)
    validate_preseed_projects(client, preseed_cfg)


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot manipulate ZFS"
)
@pytest.mark.user_data(STORAGE_USER_DATA.format("zfs"))
def test_storage_zfs(client):
    validate_storage(client, "zfsutils-linux", "zpool")


@pytest.mark.skipif(
    PLATFORM == "lxd_container", reason="Containers cannot manipulate LXD"
)
@pytest.mark.skipif(
    CURRENT_RELEASE < FOCAL, reason="Tested on focal and later"
)
def test_storage_preseed_zfs(setup_image, session_cloud: IntegrationCloud):
    # TODO: If test is marked as not bionic, why is there a bionic section?
    if CURRENT_RELEASE.series in ("bionic",):
        nictype = "nictype: bridged"
        parent = "parent: lxdbr0"
        network = ""
    else:
        nictype = ""
        parent = ""
        network = "network: lxdbr0"
    user_data = STORAGE_PRESEED_USER_DATA.format(
        driver="zfs", nictype=nictype, parent=parent, network=network
    )
    with session_cloud.launch(user_data=user_data) as client:
        validate_storage(client, "zfsutils-linux", "zpool")
        src_cfg = yaml.safe_load(user_data)
        preseed_cfg = yaml.safe_load(src_cfg["lxd"]["preseed"])
        validate_preseed_profiles(client, preseed_cfg)
        validate_preseed_storage_pools(client, preseed_cfg)
        validate_preseed_projects(client, preseed_cfg)



# See https://docs.pytest.org/en/stable/example
# /parametrize.html#parametrizing-conditional-raising

import textwrap
from unittest import mock

import pytest

from cloudinit import net
from cloudinit import safeyaml as yaml
from cloudinit.distros.networking import (
    BSDNetworking,
    LinuxNetworking,
    Networking,
)
from tests.unittests.helpers import does_not_raise, readResource


@pytest.fixture
def generic_networking_cls():
    """Returns a direct Networking subclass which errors on /sys usage.

    This enables the direct testing of functionality only present on the
    ``Networking`` super-class, and provides a check on accidentally using /sys
    in that context.
    """

    class TestNetworking(Networking):
        def apply_network_config_names(self, *args, **kwargs):
            raise NotImplementedError

        def is_physical(self, *args, **kwargs):
            raise NotImplementedError

        def settle(self, *args, **kwargs):
            raise NotImplementedError

        def try_set_link_up(self, *args, **kwargs):
            raise NotImplementedError

    error = AssertionError("Unexpectedly used /sys in generic networking code")
    with mock.patch(
        "cloudinit.net.get_sys_class_path",
        side_effect=error,
    ):
        yield TestNetworking


@pytest.fixture
def bsd_networking_cls(asset="netinfo/freebsd-ifconfig-output"):
    """Returns a patched BSDNetworking class which already comes pre-loaded
    with output for ``ifconfig -a``"""
    ifs_txt = readResource(asset)
    with mock.patch(
        "cloudinit.distros.networking.subp.subp", return_value=(ifs_txt, None)
    ):
        yield BSDNetworking


@pytest.fixture
def sys_class_net(tmpdir):
    sys_class_net_path = tmpdir.join("sys/class/net")
    sys_class_net_path.ensure_dir()
    with mock.patch(
        "cloudinit.net.get_sys_class_path",
        return_value=sys_class_net_path.strpath + "/",
    ):
        yield sys_class_net_path


class TestBSDNetworkingIsPhysical:
    def test_is_physical(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert networking.is_physical("vtnet0")

    def test_is_not_physical(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert not networking.is_physical("re0.33")


class TestBSDNetworkingIsVLAN:
    def test_is_vlan(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert networking.is_vlan("re0.33")

    def test_is_not_physical(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert not networking.is_vlan("vtnet0")


class TestBSDNetworkingIsBridge:
    def test_is_vlan(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert networking.is_bridge("bridge0")

    def test_is_not_physical(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        assert not networking.is_bridge("vtnet0")


class TestLinuxNetworkingIsPhysical:
    def test_returns_false_by_default(self, sys_class_net):
        assert not LinuxNetworking().is_physical("eth0")

    def test_returns_false_if_devname_exists_but_not_physical(
        self, sys_class_net
    ):
        devname = "eth0"
        sys_class_net.join(devname).mkdir()
        assert not LinuxNetworking().is_physical(devname)

    def test_returns_true_if_device_is_physical(self, sys_class_net):
        devname = "eth0"
        device_dir = sys_class_net.join(devname)
        device_dir.mkdir()
        device_dir.join("device").write("")

        assert LinuxNetworking().is_physical(devname)


@mock.patch("cloudinit.distros.networking.BSDNetworking.is_up")
class TestBSDNetworkingTrySetLinkUp:
    def test_calls_subp_return_true(self, m_is_up, bsd_networking_cls):
        devname = "vtnet0"
        networking = bsd_networking_cls()
        m_is_up.return_value = True

        with mock.patch("cloudinit.subp.subp") as m_subp:
            is_success = networking.try_set_link_up(devname)
            assert (
                mock.call(["ifconfig", devname, "up"])
                == m_subp.call_args_list[-1]
            )
        assert is_success


@mock.patch("cloudinit.net.is_up")
@mock.patch("cloudinit.distros.networking.subp.subp")
class TestLinuxNetworkingTrySetLinkUp:
    def test_calls_subp_return_true(self, m_subp, m_is_up):
        devname = "eth0"
        m_is_up.return_value = True
        is_success = LinuxNetworking().try_set_link_up(devname)

        assert (
            mock.call(["ip", "link", "set", devname, "up"])
            == m_subp.call_args_list[-1]
        )
        assert is_success

    def test_calls_subp_return_false(self, m_subp, m_is_up):
        devname = "eth0"
        m_is_up.return_value = False
        is_success = LinuxNetworking().try_set_link_up(devname)

        assert (
            mock.call(["ip", "link", "set", devname, "up"])
            == m_subp.call_args_list[-1]
        )
        assert not is_success


class TestBSDNetworkingSettle:
    def test_settle_doesnt_error(self, bsd_networking_cls):
        networking = bsd_networking_cls()
        networking.settle()


@pytest.mark.usefixtures("sys_class_net")
@mock.patch("cloudinit.distros.networking.util.udevadm_settle", autospec=True)
class TestLinuxNetworkingSettle:
    def test_no_arguments(self, m_udevadm_settle):
        LinuxNetworking().settle()

        assert [mock.call(exists=None)] == m_udevadm_settle.call_args_list

    def test_exists_argument(self, m_udevadm_settle):
        LinuxNetworking().settle(exists="ens3")

        expected_path = net.sys_dev_path("ens3")
        assert [
            mock.call(exists=expected_path)
        ] == m_udevadm_settle.call_args_list


class TestNetworkingWaitForPhysDevs:
    @pytest.fixture
    def wait_for_physdevs_netcfg(self):
        """This config is shared across all the tests in this class."""

        def ethernet(mac, name, driver=None, device_id=None):
            v2_cfg = {"set-name": name, "match": {"macaddress": mac}}
            if driver:
                v2_cfg["match"].update({"driver": driver})
            if device_id:
                v2_cfg["match"].update({"device_id": device_id})

            return v2_cfg

        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", "virtio", "0x1000"],
            ["00:11:22:33:44:55", "ens3", "e1000", "0x1643"],
        ]
        netcfg = {
            "version": 2,
            "ethernets": {args[1]: ethernet(*args) for args in physdevs},
        }
        return netcfg

    def test_skips_settle_if_all_present(
        self,
        generic_networking_cls,
        wait_for_physdevs_netcfg,
    ):
        networking = generic_networking_cls()
        with mock.patch.object(
            networking, "get_interfaces_by_mac"
        ) as m_get_interfaces_by_mac:
            m_get_interfaces_by_mac.side_effect = iter(
                [{"aa:bb:cc:dd:ee:ff": "eth0", "00:11:22:33:44:55": "ens3"}]
            )
            with mock.patch.object(
                networking, "settle", autospec=True
            ) as m_settle:
                networking.wait_for_physdevs(wait_for_physdevs_netcfg)
            assert 0 == m_settle.call_count

    def test_calls_udev_settle_on_missing(
        self,
        generic_networking_cls,
        wait_for_physdevs_netcfg,
    ):
        networking = generic_networking_cls()
        with mock.patch.object(
            networking, "get_interfaces_by_mac"
        ) as m_get_interfaces_by_mac:
            m_get_interfaces_by_mac.side_effect = iter(
                [
                    {
                        "aa:bb:cc:dd:ee:ff": "eth0"
                    },  # first call ens3 is missing
                    {
                        "aa:bb:cc:dd:ee:ff": "eth0",
                        "00:11:22:33:44:55": "ens3",
                    },  # second call has both
                ]
            )
            with mock.patch.object(
                networking, "settle", autospec=True
            ) as m_settle:
                networking.wait_for_physdevs(wait_for_physdevs_netcfg)
            m_settle.assert_called_with(exists="ens3")

    @pytest.mark.parametrize(
        "strict,expectation",
        [(True, pytest.raises(RuntimeError)), (False, does_not_raise())],
    )
    def test_retrying_and_strict_behaviour(
        self,
        strict,
        expectation,
        generic_networking_cls,
        wait_for_physdevs_netcfg,
    ):
        networking = generic_networking_cls()
        with mock.patch.object(
            networking, "get_interfaces_by_mac"
        ) as m_get_interfaces_by_mac:
            m_get_interfaces_by_mac.return_value = {}

            with mock.patch.object(
                networking, "settle", autospec=True
            ) as m_settle:
                with expectation:
                    networking.wait_for_physdevs(
                        wait_for_physdevs_netcfg, strict=strict
                    )

        assert (
            5 * len(wait_for_physdevs_netcfg["ethernets"])
            == m_settle.call_count
        )


class TestLinuxNetworkingApplyNetworkCfgNames:
    V1_CONFIG = textwrap.dedent(
        """\
        version: 1
        config:
            - type: physical
              name: interface0
              mac_address: "52:54:00:12:34:00"
              subnets:
                  - type: static
                    address: 10.0.2.15
                    netmask: 255.255.255.0
                    gateway: 10.0.2.2
    """
    )
    V2_CONFIG = textwrap.dedent(
        """\
      version: 2
      ethernets:
          interface0:
            match:
              macaddress: "52:54:00:12:34:00"
            addresses:
              - 10.0.2.15/24
            gateway4: 10.0.2.2
            set-name: interface0
    """
    )

    V2_CONFIG_NO_SETNAME = textwrap.dedent(
        """\
      version: 2
      ethernets:
          interface0:
            match:
              macaddress: "52:54:00:12:34:00"
            addresses:
              - 10.0.2.15/24
            gateway4: 10.0.2.2
    """
    )

    V2_CONFIG_NO_MAC = textwrap.dedent(
        """\
      version: 2
      ethernets:
          interface0:
            match:
              driver: virtio-net
            addresses:
              - 10.0.2.15/24
            gateway4: 10.0.2.2
            set-name: interface0
    """
    )

    @pytest.mark.parametrize(
        ["config_attr"],
        [
            pytest.param("V1_CONFIG", id="v1"),
            pytest.param("V2_CONFIG", id="v2"),
        ],
    )
    @mock.patch("cloudinit.net.device_devid")
    @mock.patch("cloudinit.net.device_driver")
    def test_apply_renames(
        self,
        m_device_driver,
        m_device_devid,
        config_attr: str,
    ):
        networking = LinuxNetworking()
        m_device_driver.return_value = "virtio_net"
        m_device_devid.return_value = "0x15d8"
        netcfg = yaml.load(getattr(self, config_attr))

        with mock.patch.object(
            networking, "_rename_interfaces"
        ) as m_rename_interfaces:
            networking.apply_network_config_names(netcfg)

        assert (
            mock.call(
                [["52:54:00:12:34:00", "interface0", "virtio_net", "0x15d8"]]
            )
            == m_rename_interfaces.call_args_list[-1]
        )

    @pytest.mark.parametrize(
        ["config_attr"],
        [
            pytest.param("V2_CONFIG_NO_SETNAME", id="without_setname"),
            pytest.param("V2_CONFIG_NO_MAC", id="without_mac"),
        ],
    )
    def test_apply_v2_renames_skips_without_setname_or_mac(
        self, config_attr: str
    ):
        networking = LinuxNetworking()
        netcfg = yaml.load(getattr(self, config_attr))
        with mock.patch.object(
            networking, "_rename_interfaces"
        ) as m_rename_interfaces:
            networking.apply_network_config_names(netcfg)
        m_rename_interfaces.assert_called_with([])

    def test_apply_v2_renames_raises_runtime_error_on_unknown_version(self):
        networking = LinuxNetworking()
        with pytest.raises(RuntimeError):
            networking.apply_network_config_names(yaml.load("version: 3"))
